# Copyright (c) 2021-2023 Mario S. Könz; License: MIT
from ._15_pytest import PytestMixin


class CoverageMixin(PytestMixin):
    def templated(self, negative_default: bool = False) -> None:
        super().templated(negative_default)
        self.auxcon.dependencies.test.append(self.versions.pytest_cov)



# Copyright (c) 2021-2023 Mario S. Könz; License: MIT
import itertools
import typing as tp

from .._logging import logger
from .._parser import Jinja2Parser
from .._parser import YamlParser
from .._proto_namespace import _ProtoNamespace
from ._03_meta import MetaMixin
from ._05_project import ProjectMixin
from ._06_dependency import DependencyMixin
from ._18_payload import PayloadMixin


class DockerMixin(PayloadMixin, DependencyMixin, ProjectMixin, MetaMixin):
    @classmethod
    def __keys(cls) -> tp.Tuple[str, ...]:
        return ("base_match", "platform", "compose_version")

    def templated(self, negative_default: bool = False) -> None:
        super().templated(negative_default)
        data = self.auxcon.payload

        data.setdefault("python", _ProtoNamespace())
        data.python["check-release-notes"] = _ProtoNamespace()

        if self.is_enabled("pre-commit"):
            data.setdefault("with_dependency", _ProtoNamespace())
            run = "pre-commit"
            build = f"build-{run}"
            data.with_dependency[build] = dict(deps=["build-python-deps"])
            data.with_dependency[run] = dict(variant={"all": dict(cmd="--all")})

        if self.is_enabled("pytest"):
            data.setdefault("with_dependency", _ProtoNamespace())
            run = "pytest"
            build = f"build-{run}"
            data.with_dependency[build] = dict(deps=["build-python-deps"])
            data.with_dependency[run] = dict()
            if self.is_enabled("coverage"):
                data.with_dependency[run]["variant"] = {
                    "mr": {"marker": ""},
                    "cov": dict(coverage=95, variant={"mr": None}),
                }

        if self.is_enabled("docs"):
            run = "docs"
            build = f"build-{run}"
            data.setdefault("with_dependency", _ProtoNamespace())
            data.with_dependency[build] = dict(
                deps=["build-python-deps"], param=dict(extra_req="docs")
            )
            data.with_dependency[run] = None

        if self.is_enabled("gitlab"):
            data.setdefault("docker_run", _ProtoNamespace())
            data.docker_run["gitlab-release"] = _ProtoNamespace(
                param=_ProtoNamespace(
                    environment=_ProtoNamespace(
                        RELEASE_TAG="{version}", RELEASE_DESCRIPTION="{release_note}"
                    )
                )
            )
            data.docker_run["pkg-gitlab"] = _ProtoNamespace()

    def demodata(self) -> None:
        super().demodata()
        data = self.auxcon.payload
        data.docker_settings = _ProtoNamespace(
            platform="amd64", base_match={"mybase": dict(fallback="develop")}
        )

    def update_to_template(self, tpl: _ProtoNamespace, full: _ProtoNamespace) -> None:
        super().update_to_template(tpl, full)

        for flavor in ["docker_build", "docker_run", "python", "with_dependency"]:
            if flavor not in tpl.payload:
                continue
            data = self.auxf.payload.get(flavor, _ProtoNamespace())
            old = list(data)
            new = list(tpl.payload[flavor])
            last_idx = 0
            for name, payload in full.payload[flavor].items():
                if name in new and name not in old:
                    data[name] = payload
                    old.insert(last_idx + 1, name)
                    self._print(f"payload.{flavor}: added {name}", fg="green")
                elif name in old and name not in new:
                    del data[name]
                    old.remove(name)
                    self._print(f"payload.{flavor}: removed {name}", fg="red")
                if name in old and name in new:
                    last_idx = old.index(name)
            if data:
                self.auxf.payload[flavor] = data

    def formatted(self) -> None:
        super().formatted()
        self._copy_keys_over(self.__keys(), "payload", "docker_settings")
        self._to_proto_ns("payload", "docker_settings")
        self._to_proto_ns("payload", "docker_settings", "base_match", iter_mapping=True)

    def defaulted(self) -> None:
        super().defaulted()
        payload_data = self.auxd.payload
        payload_data.setdefault("docker_settings", _ProtoNamespace())
        data = payload_data.docker_settings
        data.setdefault("base_match", _ProtoNamespace())
        data.setdefault("platform", None)
        data.setdefault("compose_version", self.versions.docker_compose_file)

    def enriched(self) -> None:
        super().enriched()
        data = self.auxe.payload.docker_settings
        data.project_dir = "../.."
        data.source_dir = f"../../{self.auxe.project.source_dir}"
        self._docker_build_enrich()
        self._docker_run_enrich()

    def _docker_build_enrich(self) -> None:
        extra_req_default = {
            "python-deps": ["default"],
            "pytest": ["test"],
            "pre-commit": ["test", "dev"],
            "pytest-standalone": ["default", "test"],
            "docs": ["docs"],
            "ansible-deploy": ["deploy"],
            "devenv": ["test", "dev"],
        }
        deps = self.auxe.dependencies
        data = self.auxe.payload
        prefix = "build-"
        for key, payload_cfg in data.docker_build.items():
            assert key.startswith(prefix)
            name = key[len(prefix) :]
            payload_cfg.setdefault("param", _ProtoNamespace())
            assert isinstance(payload_cfg.param, _ProtoNamespace)
            param = payload_cfg.param
            if "extra_req" in param and isinstance(param.extra_req, str):
                param.extra_req = [param.extra_req]

            param.setdefault("pip_req", [])
            param.setdefault("script", [])
            param.setdefault("base", None)
            self._adjust_base_on_match_entry(param, self.auxe)

            param.setdefault("apt_req", {})

            self._set_name_and_version(param, name)

            if "mode" in param:
                supported = ["django", "django+nginx", "ansible"]
                if param.mode not in supported:
                    raise RuntimeError(
                        f"mode {param.mode} is not supported {supported}"
                    )

            # branch matching and dep settings
            fallback = extra_req_default.get(param.docker_name, [])
            needed = param.get("extra_req", fallback)
            param.pip_req = self._unique_sum(
                param.pip_req, *[deps.get(x, []) for x in needed]
            )
            param.apt_req = self._unique_sum_dict(
                param.apt_req, *[deps.get(x + "_apt", {}) for x in needed]
            )
            param.script = self._sum(
                param.script, *[deps.get(x + "_script", []) for x in needed]
            )
            assert all('"' not in x for x in param.pip_req)

            if self.is_enabled("pip"):
                self.branch_match_and_cred_passing(param, self.auxe)  # type: ignore

    def _docker_run_enrich(self) -> None:
        data = self.auxe.payload
        for key, payload_cfg in data.docker_run.items():
            payload_cfg.setdefault("param", _ProtoNamespace())
            param = payload_cfg.param

            res = []
            if "assets" in param:
                if isinstance(param.assets, str):
                    param.assets = [param.assets]

                for x in param.assets:
                    shortname = x.rsplit("/", 1)[1]
                    url = "$CI_API_V4_URL/projects/$CI_PROJECT_ID/packages/generic/" + x
                    res.append(
                        r"{\"name\":\"" + shortname + r"\",\"url\":\"" + url + r"\"}"
                    )
            param.assets = res
            self._set_name_and_version(param, key)
            param.setdefault("services", [param.service_name])

    def _set_name_and_version(self, param: _ProtoNamespace, name: str) -> None:
        slug = self.auxe.project.slug
        param.service_name = name
        if param.service_name == slug:
            raise RuntimeError(
                f"service name should not be identical to project slug '{slug}'"
            )

        param.setdefault("image_name", f"{slug}-{param.service_name}")

        param.setdefault("docker_name", name)
        param.setdefault("version", self.auxe.python_project.minimal_version)
        if not isinstance(param.version, str):
            raise RuntimeError(
                f"please specity the version '{param.version}' in {name} as a string"
            )
        param.setdefault("version_slug", param.version.replace(".", ""))

    def _adjust_base_on_match_entry(
        self,
        opts: _ProtoNamespace,
        auxe: _ProtoNamespace,
    ) -> None:
        base_match = auxe.payload.docker_settings.base_match
        if opts.base is None:
            return
        image, tag = opts.base.rsplit(":", 1)
        if image not in base_match:
            return
        fallback = base_match[image].fallback
        var = (
            f'{image.upper().replace("/", "_").replace(".", "_").replace("-", "_")}_TAG'
        )
        skip = [auxe.gitlab.release_branch]

        # changing this order affects the jinja2 files!
        opts.base_match = [(image, var, skip, fallback, tag)]

        opts.base = f"{image}:${var}"

    def bake(self) -> None:  # pylint: disable=too-many-branches
        super().bake()
        data = self.auxe.payload
        settings = self.auxe.payload.docker_settings

        config = _ProtoNamespace(
            [("version", settings.compose_version), ("services", {})]
        )

        # pylint: disable=too-many-nested-blocks
        for val in data.docker_run.values():
            self._bake_docker_compose_file(val, config)

        for val in data.docker_build.values():
            self._bake_docker_build_file(val)
            self._bake_docker_compose_file(val, config)

        dest = self.target / "docker/compose.yml"
        written = YamlParser.write(config, dest)
        if written:
            self._print(f"baked {dest}", fg="green")

        # self._add_compositions(config)

    def _bake_docker_build_file(self, payload: _ProtoNamespace) -> None:
        opts = payload.param

        for name, from_adaux in [
            (f"docker/{opts.docker_name}.dockerfile", False),
            (f"docker/{opts.docker_name}/Dockerfile", False),
            (f"docker/services/{opts.docker_name}/Dockerfile", True),
        ]:
            try:
                self.bake_file(
                    name,
                    f"docker/{opts.service_name}.dockerfile",
                    opts=opts,
                    custom=not from_adaux,
                )
                if from_adaux:
                    logger.debug(
                        "%s: Dockerfile found@%s [adaux]", opts.docker_name, name
                    )
                else:
                    logger.info("%s: Dockerfile found@%s", opts.docker_name, name)
                return
            except FileNotFoundError:
                logger.debug("%s: fail@%s", opts.docker_name, name)
        raise RuntimeError(
            f"Dockerfile for {opts.docker_name} not found. "
            "Run this command with -vvv to see valid file locations"
        )

    def _bake_docker_compose_file(
        self, payload: _ProtoNamespace, config: _ProtoNamespace
    ) -> None:
        # pylint: disable=too-many-locals
        opts = payload.param

        input_config, _ = self._get_part_config(opts)
        valid = {"services", "networks"}
        invalid_keys = set(input_config.keys()) - valid
        if invalid_keys:
            raise RuntimeError(
                f"only {valid} can be defined in compose files. Found {invalid_keys}"
            )

        for valid, (key, val) in itertools.product(  # bc for-for-break
            {opts.service_name, opts.docker_name},  # check service first
            input_config.services.items(),
        ):
            if key == valid:
                services = _ProtoNamespace({opts.service_name: val})
                self._update_config(payload, services, config)
                break
        else:
            raise NotImplementedError(opts.service_name)

        # update networks
        config.setdefault("networks", _ProtoNamespace())
        custom_networks = input_config.get("networks", _ProtoNamespace())
        for network, val in custom_networks.items():
            if network in config["networks"]:
                if config["networks"][network] != val:
                    raise RuntimeError(
                        f"network {network} already defined differently..."
                    )
        config["networks"].update(custom_networks)
        if not config["networks"]:
            del config["networks"]

    def _get_part_config(
        self, opts: _ProtoNamespace
    ) -> tp.Tuple[_ProtoNamespace, bool]:
        logger.debug(
            "%s: looking for compose info %s...",
            opts.service_name,
            opts.docker_name,
        )
        src_dir = self.root / "docker" / "services"
        custom_dir = self.target_custom / "docker"

        # service could be a variant with no own docker name
        # docker_name must exist
        for path, from_adaux in [
            (custom_dir / opts.service_name / "compose", False),
            (custom_dir / opts.service_name, False),
            (custom_dir / opts.docker_name / "compose", False),
            (custom_dir / opts.docker_name, False),
            (src_dir / opts.docker_name / "compose", True),
            (custom_dir / "compose", False),
        ]:
            src = path.with_suffix(".yml")
            srcj = path.with_suffix(".yml.jinja2")
            if srcj.exists():
                with Jinja2Parser.render_to_tmp(srcj, aux=self.auxe, opts=opts) as src:
                    logger.info("%s: found@%s", opts.docker_name, srcj)
                    return YamlParser.read(src), from_adaux

            if src.exists():
                logger.info("%s: found@%s", opts.docker_name, src)
                return YamlParser.read(src), from_adaux
            logger.debug("%s: fail@%s[.jinja2]", opts.docker_name, src)

        raise RuntimeError(
            f"Definition for {opts.docker_name} not found. "
            "Run this command with -vvv to see valid file locations"
        )

    def _update_config(
        self,
        payload: _ProtoNamespace,
        custom_services: _ProtoNamespace,
        config: _ProtoNamespace,
    ) -> None:
        opts = payload.param

        if (
            opts.docker_name not in custom_services
            and opts.docker_name not in config["services"]
        ):
            if opts.services == [opts.service_name]:
                raise RuntimeError(f"could not find service {opts.docker_name}")
            logger.debug(
                "%s found no match, but is accepted as a collection of services: %s",
                opts.docker_name,
                opts.services,
            )

        for service_name, val in custom_services.items():
            for key in list(val):
                if payload.flavor == "docker_build":
                    if key not in ["build", "image"]:
                        del val[key]
                elif payload.flavor == "docker_run":
                    if key in ["build"]:
                        del val[key]
                else:
                    raise NotImplementedError(payload.flavor)
            logger.debug("updating config with %s", service_name)
            config["services"].setdefault(service_name, val)
            config["services"][service_name].update(val)

    @staticmethod
    def _unique_sum(*args: tp.List[str]) -> tp.List[str]:
        res = []
        for part in args:
            for x in part:
                if x not in res:
                    res.append(x)
        return res

    @staticmethod
    def _unique_sum_dict(
        *args: tp.Dict[str, tp.List[str]]
    ) -> tp.Dict[str, tp.List[str]]:
        res: tp.Dict[str, tp.List[str]] = {}
        for part in args:
            for key, val in part.items():
                res.setdefault(key, [])
                for x in val:
                    if x not in res[key]:
                        res[key].append(x)

        return res

    @staticmethod
    def _sum(*args: tp.List[str]) -> tp.List[str]:
        res = []
        for part in args:
            for x in part:
                res.append(x)
        return res



# Copyright (c) 2021-2023 Mario S. Könz; License: MIT
import os
import time

from ._cli import adaux
from ._cli import click
from ._cli import convert_runtime_to_click_error


@adaux.group()  # type: ignore[unused-ignore,misc]
@click.pass_context
@click.option(
    "-t",
    "--token",
    prompt=os.environ.get("GITLAB_API_TOKEN", "") == "",
    help="Gitlab Access Token with API Scope.",
    default="env:GITLAB_API_TOKEN",
)
def gitlab(ctx: click.Context, token: str) -> None:
    ctx.obj.token = _get_token_from_env(token)


@gitlab.command()  # type: ignore[unused-ignore,misc]
@click.pass_context
def init(ctx: click.Context) -> None:
    r"""
    Sets various settings on the remote gitlab repositroy.
    """
    with convert_runtime_to_click_error(ctx):
        ctx.obj.gitlab(ctx.obj.token)


@gitlab.command()  # type: ignore[unused-ignore,misc]
@click.pass_context
@click.option(
    "-w",
    "--watch",
    default=False,
    is_flag=True,
    help="Check the pipeline status all 30 seconds.",
)
@click.option(
    "-a",
    "--show-success",
    is_flag=True,
    default=False,
    help="Show successful jobs.",
)
def pipeline(ctx: click.Context, watch: bool, show_success: bool) -> None:
    r"""
    Shows the most recend pipeline on the cli.
    """
    with convert_runtime_to_click_error(ctx):
        while True:
            done, _ = ctx.obj.pipeline(ctx.obj.token, show_success)
            if done or not watch:
                break
            time.sleep(30)


@gitlab.command()  # type: ignore[unused-ignore,misc]
@click.pass_context
@click.option(
    "-w",
    "--watch",
    default=False,
    is_flag=True,
    help="Check the release status all 30 seconds.",
)
@click.option(
    "-p",
    "--pipeline-check",
    default=False,
    is_flag=True,
    help="Wait for successful pipeline status before executing release.",
)
@click.option(
    "-r",
    "--rerelease",
    default=False,
    is_flag=True,
    help="Allow the MR to have the same name as past closed MR.",
)
def release(  # pylint: disable=too-many-arguments
    ctx: click.Context, watch: bool, pipeline_check: bool, rerelease: bool
) -> None:
    r"""
    Create and display release merge request.
    """
    with convert_runtime_to_click_error(ctx):
        success = not pipeline_check
        if pipeline_check:
            while True:
                done, success = ctx.obj.pipeline(ctx.obj.token, show_success=False)
                if done or not watch:
                    break
                time.sleep(30)

        # will be skipped if pipeline is checked and failed
        while success:
            done, _ = ctx.obj.release(ctx.obj.token, rerelease)
            if done or not watch:
                break
            time.sleep(30)


@gitlab.command()  # type: ignore[unused-ignore,misc]
@click.pass_context
@click.argument("mr_message", nargs=-1)
def mr(ctx: click.Context, mr_message: str) -> None:
    """
    Create a merge request from the current branch to the default branch.
    """
    with convert_runtime_to_click_error(ctx):
        ctx.obj.merge_request(ctx.obj.token, " ".join(mr_message))


def _get_token_from_env(token: str) -> str:
    default = os.environ.get("GITLAB_API_TOKEN", "")
    if default == "":
        if token == "env:GITLAB_API_TOKEN":
            raise click.UsageError("GITLAB_API_TOKEN is not set!")
        click.secho(
            f"if you want to avoid re-entering, run: ' export GITLAB_API_TOKEN={token}'",
            fg="yellow",
        )

    if token == "env:GITLAB_API_TOKEN":
        token = default
    return token



# Copyright (c) 2021-2023 Mario S. Könz; License: MIT
import contextlib
import copy
import dataclasses as dc
import os
import typing as tp

from ..._logging import logger
from ..._proto_namespace import _ProtoNamespace

__all__ = ["Payload"]


@dc.dataclass
class Payload:
    flavor: tp.ClassVar[str]
    name: str
    auxh: _ProtoNamespace
    environment: tp.Dict[str, str]

    def run(self, force: bool) -> None:  # pylint: disable=unused-argument
        raise NotImplementedError()

    def is_up_to_date(self) -> bool:
        return False

    def hydrate(self, deps: tp.Tuple["Payload", ...] = tuple()) -> None:
        pass

    @contextlib.contextmanager
    def write_environment_variables(
        self, param: _ProtoNamespace
    ) -> tp.Iterator[tp.Dict[str, str]]:
        param = copy.copy(param)
        old_env = {}
        for key, val in param.pop("environment", {}).items():
            old_env[key] = os.environ.get(key, None)
            os.environ[key] = self.auxh.env_var.apply_format(val)
            from_tpl = ""
            if os.environ[key] != val:
                from_tpl = f" <- {val}"
            logger.info(
                "payload %s env: %s = %s%s", self.name, key, os.environ[key], from_tpl
            )
        yield param
        for key, val in old_env.items():
            if val is None:
                del os.environ[key]
                logger.info("payload %s env: %s [removed]", self.name, key)
            else:
                os.environ[key] = val
                logger.info("payload %s env: %s = %s [restored]", self.name, key, val)



# Copyright (c) 2021-2023 Mario S. Könz; License: MIT
import typing as tp

from .._proto_namespace import _ProtoNamespace
from ._05_python_project import PythonProjectMixin


class PackageMixin(PythonProjectMixin):
    @classmethod
    def __keys(cls) -> tp.Tuple[str, ...]:
        return ("include", "exclude")

    def formatted(self) -> None:
        super().formatted()
        self._copy_keys_over(self.__keys(), "package")
        for key in self.__keys():
            self._to_list("package", key)

    def defaulted(self) -> None:
        super().defaulted()
        self.auxd.setdefault("package", _ProtoNamespace())
        for key in self.__keys():
            self.auxd.package.setdefault(key, [])

    def bake(self) -> None:
        super().bake()
        config = self.auxe.python_project.config
        data = self.auxe.package
        name = self.auxe.project.name

        for dkey, ckey in [
            ("include", "options.package_data"),
            ("exclude", "options.exclude_package_data"),
        ]:
            if not data[dkey]:
                continue
            config.setdefault(ckey, _ProtoNamespace())
            config[ckey][name] = data[dkey]



# automatically generated by auxilium
import contextlib
import sys
import typing as tp
from pathlib import Path

from bs4 import BeautifulSoup


def main() -> None:
    filename = Path(sys.argv[1])

    with modify_html_file(filename) as soup:
        rename_external_link(soup)
        switch_internal_external_link(soup)
        remove_private_module_parts(soup)


@contextlib.contextmanager
def modify_html_file(filename: Path) -> tp.Iterator[BeautifulSoup]:
    with open(filename, "rb") as f:
        soup = BeautifulSoup(f, "html.parser")
    yield soup
    with open(filename, "wb") as f:
        f.write(soup.encode(formatter="html"))


def rename_external_link(soup: BeautifulSoup) -> None:
    for x in soup.select(".reference.external .viewcode-link .pre"):
        x.string = "[git]"


def switch_internal_external_link(soup: BeautifulSoup) -> None:
    for x in soup.select(".sig.sig-object.py"):
        to_add = list(x.children)
        external = to_add[-3]
        internal = to_add[-2]

        if "internal" in internal["class"] and "external" in external["class"]:
            to_add[-3] = internal
            to_add[-2] = external

            x.clear()

            for i, y in enumerate(to_add):
                x.insert(i, y)


def remove_private_module_parts(soup: BeautifulSoup) -> None:
    for x in soup.select(".reference.internal"):
        title = x.get("title")
        if not title:
            continue

        if "._" in x.string:
            x.string = ".".join(x for x in x.string.split(".") if not x.startswith("_"))
            x["title"] = x.string

    for x in soup.select(".reference.external"):
        title = x.get("title")
        if not title:
            continue
        if "._" in x.string:
            x.string = ".".join(x for x in x.string.split(".") if not x.startswith("_"))


if __name__ == "__main__":
    main()



# Copyright (c) 2021-2023 Mario S. Könz; License: MIT
import copy
import os
import subprocess
import sys
import typing as tp
from pathlib import Path

from .._parser import ConfigParser
from .._proto_namespace import _ProtoNamespace
from ._03_meta import MetaMixin
from ._06_dependency import DependencyMixin


class PylintMixin(DependencyMixin, MetaMixin):
    @classmethod
    def __keys(cls) -> tp.Tuple[str, ...]:
        return (
            "disable",
            "good_names",
            "class_attr_rgx",
            "min_similarity_lines",
            "argument_rgx",
            "attr_rgx",
            "function_rgx",
            "variable_rgx",
            "method_rgx",
        )

    def templated(self, negative_default: bool = False) -> None:
        super().templated(negative_default)
        data = self.auxcon.pylint = _ProtoNamespace()
        self.auxcon.dependencies.dev.append(self.versions.pylint)
        data.disable = [
            "line-too-long",
            "wildcard-import",
            "use-dict-literal",
            "wrong-import-position",
            "too-few-public-methods",
        ]
        if not self.is_enabled("docs") or negative_default:
            data.disable += [
                "missing-class-docstring",
                "missing-module-docstring",
                "missing-function-docstring",
            ]

    def update_to_template(self, tpl: _ProtoNamespace, full: _ProtoNamespace) -> None:
        super().update_to_template(tpl, full)
        data = self.auxf.pylint
        data_full = full.pylint
        data_tpl = tpl.pylint
        # will be left alone once defined, not removed
        hysteresis = [
            "missing-class-docstring",
            "missing-module-docstring",
            "missing-function-docstring",
        ]

        old_disable = data.disable
        order = data_tpl.disable + old_disable + hysteresis

        def ordered(list_: tp.Iterable[str]) -> tp.List[str]:
            return list(sorted(list_, key=order.index))

        should_be_present = set(data_tpl.disable)
        custom = set(old_disable) - (set(data_full.disable) - set(hysteresis))

        data.disable = ordered(should_be_present | custom)

        added = set(data.disable) - set(old_disable)
        removed = set(old_disable) - set(data.disable)
        if added:
            self._print(f"pylint.disable: added {ordered(added)}", fg="green")
        if removed:
            self._print(f"pylint.disable: removed {ordered(removed)}", fg="red")

    def demodata(self) -> None:
        super().demodata()
        data = self.auxcon.pylint
        data.setdefault("disable", [])
        data.disable += ["too-few-public-methods", "no-self-use"]
        data.good_names = ["t", "dt"]

    def formatted(self) -> None:
        super().formatted()
        self._copy_keys_over(self.__keys(), "pylint")
        self._copy_keys_over(self.__keys(), "pylint_test")
        for key in self.__keys()[:2]:
            self._to_list("pylint", key)
            self._to_list("pylint_test", key)

    def defaulted(self) -> None:
        super().defaulted()
        self.auxd.setdefault("pylint", _ProtoNamespace())
        self.auxd.setdefault("pylint_test", _ProtoNamespace())
        for key in self.__keys()[:2]:
            self.auxd.pylint.setdefault(key, [])
            self.auxd.pylint_test.setdefault(key, [])

    def bake(self) -> None:  # pylint: disable=too-many-locals,too-many-branches
        super().bake()
        # ensure python version
        found_version = ".".join(map(str, sys.version_info[:2]))
        if found_version != self.auxe.python_project.minimal_version:
            raise RuntimeError(
                f"you are using python {found_version}, please use python {self.auxe.python_project.minimal_version} (minimal version)."
            )

        # the home of the user has a potential .pylintrc file,
        # which will influence the creation
        env = os.environ.copy()
        pylint_rc = Path(os.environ["HOME"]) / ".pylintrc"
        if pylint_rc.exists():
            env["HOME"] = "/tmp"
        resp = subprocess.run(
            ["pylint", "--generate-rcfile"], capture_output=True, check=True, env=env
        )
        text = resp.stdout.decode()

        hooks = self.auxe.pre_commit.hooks
        configs = [ConfigParser.read_string(text)]
        self.patch(configs[0])
        keys = ["pylint"]
        writeout = ["pylint" in hooks]

        writeout.append("pylint-test" in hooks)
        if writeout[-1]:
            keys.append("pylint-test")

        for i, key in enumerate(keys):
            config = configs[-1]
            if i > 0:
                config = copy.deepcopy(config)
                configs.append(config)

            key = key.replace("-", "_")
            for key1, key2, key3 in [
                ("MESSAGES CONTROL", "disable", "disable"),
                ("BASIC", "good-names", "good_names"),
                ("SIMILARITIES", "min-similarity-lines", "min_similarity_lines"),
                ("BASIC", "variable-rgx", "variable_rgx"),
                ("BASIC", "const-rgx", "const_rgx"),
                ("BASIC", "function-rgx", "function_rgx"),
                ("BASIC", "argument-rgx", "argument_rgx"),
                ("BASIC", "class-rgx", "class_rgx"),
                ("BASIC", "attr-rgx", "attr_rgx"),
                ("BASIC", "method-rgx", "method_rgx"),
                ("BASIC", "attr-rgx", "attr_rgx"),
                ("BASIC", "class-attribute-rgx", "class_attr_rgx"),
            ]:
                subconfig = config[key1]
                data = self.auxe[key]
                if key2 in ["good-names", "disable"]:
                    proplist = copy.copy(subconfig[key2])
                    for x in data[key3]:
                        proplist[-1] += ","
                        if x in proplist:
                            raise RuntimeError(f"{x} is already marked {key3}")
                        proplist.append(x)
                    subconfig[key2] = proplist
                else:
                    if key3 in data:
                        subconfig[key2] = data[key3]

            config["SIMILARITIES"]["ignore-imports"] = "yes"

        for key, config, wout in zip(keys, configs, writeout):
            if not wout:
                continue
            dest = self.target / f"pre-commit/{key}rc"
            written = ConfigParser.write(config, dest)
            if written:
                self._print(f"baked {dest}", fg="green")

    def patch(self, config: _ProtoNamespace) -> None:
        # 2023.11, due to different ordering of use-symbolic-message-instead
        # in clang and gcc pylint-3.0.2, we need to get clang to gcc
        disable: tp.List[str] = config["MESSAGES CONTROL"]["disable"]
        move_obj = "use-symbolic-message-instead,"
        if move_obj in disable:
            move_after = "use-implicit-booleaness-not-comparison-to-zero"
            disable[disable.index(move_after)] += ","
            disable.remove(move_obj)
            disable.append(move_obj[:-1])



# Copyright (c) 2021-2023 Mario S. Könz; License: MIT
import io

from .._parser import ConfigParser
from .._proto_namespace import _ProtoNamespace
from ._08_pip import PipMixin


class EntryPointMixin(PipMixin):
    def formatted(self) -> None:
        super().formatted()
        if "entry_points" in self.auxcon:
            self.auxf.entry_points = _ProtoNamespace(self.auxcon.entry_points or {})

    def defaulted(self) -> None:
        super().defaulted()
        self.auxd.setdefault("entry_points", _ProtoNamespace())

    def enriched(self) -> None:
        super().enriched()
        stream = io.StringIO()
        ConfigParser.write_stream(self.auxe.entry_points, stream)
        self.auxe.pip.entry_points_filecontent = stream.getvalue().replace("\n", "\\n")

    def demodata(self) -> None:
        super().demodata()
        self.auxcon.entry_points = _ProtoNamespace()
        data = self.auxcon.entry_points
        data.plugin_hook = dict(plugin_name="plugin:func")

    def bake(self) -> None:
        super().bake()
        config = self.auxe.python_project.config
        data = self.auxe.entry_points
        config["options.entry_points"].update(data)



# Copyright (c) 2021-2023 Mario S. Könz; License: MIT
import typing as tp
from pathlib import Path


class FileOpsConvenience:
    @classmethod
    def ensure_parent(cls, dest: Path) -> None:
        target = dest.resolve().parent
        if not target.exists():
            target.mkdir(parents=True, exist_ok=True)


class BaseParser(FileOpsConvenience):
    @classmethod
    def rec_walk(cls, obj: tp.Any, mods: tp.Any) -> tp.Any:
        for ident, modify in mods:
            if ident(obj):
                obj = modify(obj)
        if hasattr(obj, "items"):
            for key, val in obj.items():
                obj[key] = cls.rec_walk(val, mods)
        elif isinstance(obj, list):
            obj = [cls.rec_walk(val, mods) for val in obj]
        return obj

    @classmethod
    def read(cls, filename: Path) -> tp.Any:
        raise NotImplementedError()

    @classmethod
    def write(cls, data: tp.Any, dest: Path) -> bool:
        raise NotImplementedError()



# Copyright (c) 2021-2023 Mario S. Könz; License: MIT
import dataclasses as dc
import typing as tp

from ._docker import DockerPayload
from ._docker_executors import DockerRun

__all__ = ["DockerRunPayload"]


@dc.dataclass
class DockerRunPayload(DockerPayload):
    flavor: tp.ClassVar[str] = "docker_run"

    def create_executor(self, parents: tp.Any) -> tp.Any:
        return DockerRun(
            slug=self.auxh.project.slug,
            services=self.param.services,
            parents=parents,
            images=self.param.get("images", []),
            interactive=self.param.get("interactive", False),
            exec_join=self.param.get("exec_join", False),
            run_ports=self.param.get("run_ports", []),
        )



# Copyright (c) 2021-2023 Mario S. Könz; License: MIT
import copy
import typing as tp

from .._logging import logger
from .._parser import TomlParser
from .._proto_namespace import _ProtoNamespace
from ._03_meta import MetaMixin
from ._06_dependency import DependencyMixin


class RuffMixin(DependencyMixin, MetaMixin):
    @classmethod
    def __keys(cls) -> tp.Tuple[str, ...]:
        return (
            "select",
            "ignore",
            "per_file_ignores",
        )

    def templated(self, negative_default: bool = False) -> None:
        super().templated(negative_default)
        data = self.auxcon.ruff = _ProtoNamespace()
        self.auxcon.dependencies.dev.append(self.versions.ruff)
        data.select = ["E", "F", "B"]
        data.ignore = ["E501"]
        data.per_file_ignores = {"__init__.py": ["F401", "E402", "F403", "F405"]}

    def update_to_template(self, tpl: _ProtoNamespace, full: _ProtoNamespace) -> None:
        super().update_to_template(tpl, full)
        if "ruff" not in self.auxf:
            return
        logger.warning("ruff not fully supported for sync yet")
        data = self.auxf.ruff
        data_full = full.ruff
        data_tpl = tpl.ruff

        old_ignore = data.get("ignore", [])
        order = data_tpl.ignore + old_ignore

        def ordered(list_: tp.Iterable[str]) -> tp.List[str]:
            return list(sorted(list_, key=order.index))

        should_be_present = set(data_tpl.ignore)
        custom = set(old_ignore) - set(data_full.ignore)

        data.ignore = ordered(should_be_present | custom)

        added = set(data.ignore) - set(old_ignore)
        removed = set(old_ignore) - set(data.ignore)
        if added:
            self._print(f"ruff.ignore: added {ordered(added)}", fg="green")
        if removed:
            self._print(f"ruff.ignore: removed {ordered(removed)}", fg="red")

    def formatted(self) -> None:
        super().formatted()
        self._copy_keys_over(self.__keys(), "ruff")
        self._copy_keys_over(self.__keys(), "ruff_test")
        for key in self.__keys()[:2]:
            self._to_list("ruff", key)
            self._to_list("ruff_test", key)

    def defaulted(self) -> None:
        super().defaulted()
        self.auxd.setdefault("ruff", _ProtoNamespace())
        self.auxd.setdefault("ruff_test", _ProtoNamespace())
        for key in self.__keys()[:2]:
            self.auxd.ruff.setdefault(key, [])
            self.auxd.ruff_test.setdefault(key, [])

    def bake(self) -> None:  # pylint: disable=too-many-locals,too-many-branches
        super().bake()
        if "ruff" not in self.auxf:
            return
        dest = self.target / "pre-commit/ruff.toml"
        data = copy.copy(self.auxe.ruff)
        data["cache-dir"] = "devops/cache/ruff"
        for key in ["per_file_ignores"]:
            data[key.replace("_", "-")] = data.pop(key)
        written = TomlParser.write(data, dest)
        if written:
            self._print(f"baked {dest}", fg="green")

        for key, val in self.auxf.get("ruff_test", {}).items():
            prev = data.get(key, None)
            if prev is None:
                updated = val
            if isinstance(prev, list):
                assert isinstance(val, list)
                updated = prev + val
            if isinstance(prev, dict):
                print(prev, val)
                assert isinstance(val, dict)
                updated = {**prev, **val}
            data[key] = updated

        dest = self.target / "pre-commit/ruff-test.toml"
        written = TomlParser.write(data, dest)
        if written:
            self._print(f"baked {dest}", fg="green")



# Copyright (c) 2021-2023 Mario S. Könz; License: MIT
import subprocess

from .._proto_namespace import _ProtoNamespace
from ._03_meta import MetaMixin
from ._05_project import ProjectMixin


class GitlabMixin(ProjectMixin, MetaMixin):
    def templated(self, negative_default: bool = False) -> None:
        super().templated(negative_default)
        self.auxcon.gitlab = _ProtoNamespace(
            vip_branches=_ProtoNamespace(
                develop=_ProtoNamespace(push_access_level=40, allow_force_push=True),
                release=_ProtoNamespace(),
            )
        )

    def demodata(self) -> None:
        super().demodata()
        data = self.auxcon.gitlab
        data.default_branch = "develop"
        data.release_branch = "release"
        data.remote_user = "administratum"
        data.remote_name = "auxilium"
        data.remote_url = "gitlab.x.y"
        data.squash = True

    def formatted(self) -> None:
        super().formatted()
        keys = [
            "remote_user",
            "remote_name",
            "remote_url",
            "default_branch",
            "release_branch",
            "vip_branches",
            "squash",
        ]
        self._copy_keys_over(keys, "gitlab")
        self._to_proto_ns("gitlab", "vip_branches", iter_mapping=True)

    def defaulted(self) -> None:
        super().defaulted()
        self.auxd.setdefault("gitlab", _ProtoNamespace())
        data = self.auxd.gitlab
        data.setdefault("vip_branches", _ProtoNamespace())

    def update_to_template(self, tpl: _ProtoNamespace, full: _ProtoNamespace) -> None:
        super().update_to_template(tpl, full)
        if "gitlab" not in self.auxf:
            self.auxf.gitlab = tpl.gitlab
            self._print("gitlab added", fg="green")
            return
        data = self.auxf.gitlab
        if self.is_enabled("docs"):
            if "remote_url" in data and "remote_user" in data:
                return
            proc = subprocess.run(
                ["git", "remote", "-v"], capture_output=True, check=False
            )
            if proc.returncode != 0:
                return
            full_url = proc.stdout.decode("utf-8").split()[1]
            _, url = full_url.split("@", 1)
            defaults = {}
            defaults["remote_url"], rest = url.split(":", 1)
            defaults["remote_user"], rest = rest.split("/", 1)
            defaults["remote_name"] = rest.replace(".git", "")
            key = "remote_url"
            for key, val in defaults.items():
                if key not in data:
                    if key == "remote_name" and val == self.auxf.project.name:
                        continue
                    data[key] = val
                    self._print(f"gitlab.{key}: added {data[key]}", fg="green")

    def enriched(self) -> None:
        super().enriched()
        data = self.auxe.gitlab
        vip_branch_names = list(data.vip_branches)
        defaults = dict(
            default_branch=vip_branch_names[0], release_branch=vip_branch_names[-1]
        )
        for key, val in defaults.items():
            data.setdefault(key, val)

        data.setdefault("squash", True)
        assert isinstance(data.squash, bool)
        # https://docs.gitlab.com/ee/api/protected_branches.html
        default_branch_settings = {
            (False, False): dict(
                allow_force_push=True, push_access_level=30, merge_access_level=30
            ),
            (True, False): dict(push_access_level=0, merge_access_level=30),
            (False, True): dict(push_access_level=0, merge_access_level=40),
            (True, True): dict(push_access_level=0, merge_access_level=40),
        }
        for key, val in data.vip_branches.items():
            mark = (key == data.default_branch, key == data.release_branch)
            for skey, sval in default_branch_settings[mark].items():
                val.setdefault(skey, sval)

        if self.is_enabled("pythonproject"):
            if "remote_user" in data and "remote_url" in data:
                remote_name = data.get(
                    "remote_name", self.auxe.python_project.second_name
                )
                self.auxe.python_project.project_urls.Source = (
                    f"https://{data.remote_url}/{data.remote_user}/{remote_name}"
                )

        proc = subprocess.run(
            ["git", "rev-parse", "--abbrev-ref", "HEAD"],
            capture_output=True,
            check=False,
        )
        if proc.returncode == 0:
            data.setdefault("current_branch", proc.stdout.decode("utf-8").strip())



# Copyright (c) 2021-2023 Mario S. Könz; License: MIT
from ._03_meta import MetaMixin
from ._04_migrate import MigrateMixin
from ._04_monotonic_version import MonotonicVersionMixin
from ._05_project import ProjectMixin
from ._05_python_project import PythonProjectMixin
from ._06_dependency import DependencyMixin
from ._07_package import PackageMixin
from ._08_pip import PipMixin
from ._09_gitignore import GitIgnoreMixin
from ._10_gitlab import GitlabMixin
from ._11_precommit import PrecommitMixin
from ._12a_pylint import PylintMixin
from ._12b_ruff import RuffMixin
from ._13_executable import ExecutablesMixin
from ._13b_entry_points import EntryPointMixin
from ._14_mypy import MypyMixin
from ._15_pytest import PytestMixin
from ._16_coverage import CoverageMixin
from ._17_docs import DocsMixin
from ._18_payload import PayloadMixin
from ._19_docker import DockerMixin
from ._20_ci import CiMixin
from ._21_multi_manage import MultiManageMixin
from ._98_sentinel import SentinelMixin

__all__ = ["AllComponents"]


class AllComponents(  # pylint: disable=too-many-ancestors
    SentinelMixin,
    MultiManageMixin,
    DocsMixin,
    CiMixin,
    DockerMixin,
    PayloadMixin,
    PackageMixin,
    ExecutablesMixin,
    EntryPointMixin,
    PylintMixin,
    RuffMixin,
    MypyMixin,
    PrecommitMixin,
    CoverageMixin,
    PytestMixin,
    PipMixin,
    DependencyMixin,
    GitlabMixin,
    GitIgnoreMixin,
    PythonProjectMixin,
    ProjectMixin,
    MigrateMixin,
    MonotonicVersionMixin,
    MetaMixin,
):
    pass



# Copyright (c) 2021-2023 Mario S. Könz; License: MIT
import dataclasses as dc
import datetime
import functools
import subprocess
import time
import typing as tp
import urllib.parse

from ._components import AllComponents
from ._logging import logger
from ._util import ApiRequestCommunicator

# graphql queries for image tags
# query = '{project(fullPath: "'+self.repo_name+'") { containerRepositories(name: "'+"adaux-pre-commit"+'") { nodes {id} }}}'
# resp = self.graphql_request(query=query)
# res = resp["data"]["project"]["containerRepositories"]["nodes"][0]["id"]
# query = '{containerRepository(id: "'+res+'"){tags{nodes{name}}}}'
# resp = self.graphql_request(query=query)
# print([x["name"] for x in resp["data"]["containerRepository"]["tags"]["nodes"]])


@dc.dataclass
class GitlabSetter(ApiRequestCommunicator):
    ns: AllComponents
    token: str
    repo_name: str = dc.field(init=False)
    greeting: bool = True

    def __post_init__(self) -> None:
        out = subprocess.run(["git", "remote", "-v"], capture_output=True, check=True)
        (line,) = (x.strip() for x in out.stdout.decode().split("\n") if "fetch" in x)
        line = line.rsplit(" ", 1)[0]
        host, repo = line.split("git@")[1].split(":")
        base_url = "https://" + host
        is_name = repo.replace(".git", "")
        should_name = self.ns.auxe.project.name.replace(".", "/")
        if is_name != should_name:
            logger.info("names dont match %s != %s", is_name, should_name)

        self.repo_name = is_name
        self.base_url = base_url
        # self.headers = {"Authorization": f"Bearer {self.token}"}
        self.headers = {"PRIVATE-TOKEN": self.token}
        if self.greeting:
            self._print(
                f"working on project {self.repo_name} and instance {self.base_url}",
                fg="white",
            )

    def _print(self, msg: str, **kwgs: tp.Any) -> None:
        # pylint: disable=protected-access
        self.ns._print(msg, **kwgs)

    @functools.cached_property
    def project_id(self) -> str:
        # old impl
        # query = '{project(fullPath: "' + self.repo_name + '") {id}}'
        # resp = self.graphql_request(query=query)
        # return resp["data"]["project"]["id"].split("/")[-1]  # type: ignore
        return urllib.parse.quote_plus(self.repo_name)

    def sync_settings(self) -> None:
        resp = self.api_request("projects", self.project_id)

        settings = dict(
            ci_config_path="devops/CI/00-main.yml",
            ci_default_git_depth=2,
            container_expiration_policy=dict(
                enabled=True,
                cadence="1d",
                keep_n=1,
                older_than="7d",
                name_regex_keep=None,
                name_regex="aux-.*",
            ),
            default_branch=self.ns.auxe.gitlab.default_branch,
        )

        for key, val in settings.items():
            added = ""
            outdated = self._check_outdated(val, resp[key])
            if isinstance(val, dict):
                added += "_attributes"
            if outdated:
                resp = self.api_request(
                    "projects", self.project_id, mode="put", **{key + added: val}
                )
                self._print(f"set {key} to {resp[key]}", fg="yellow")
            else:
                self._print(f"ok: {key}", fg="green")

    def sync_badges(self) -> None:
        resp = self.api_request("projects", self.project_id, "badges")
        names = [y["name"] for y in resp]

        settings = dict(
            Pipeline=dict(
                link_url=self.url(
                    "%{project_path}/-/pipelines/%{default_branch}/latest"
                ),
                image_url=self.url(
                    "%{project_path}/badges/%{default_branch}/pipeline.svg"
                ),
            ),
            Coverage=dict(
                link_url=self.url("%{project_path}/-/commits/%{default_branch}"),
                image_url=self.url(
                    "%{project_path}/badges/%{default_branch}/coverage.svg"
                ),
            ),
        )
        resp_dict = {x["name"]: x for x in resp}
        for badge_name, setting in settings.items():
            if badge_name not in names:
                msg = "installed"
                resp = self.api_request(
                    "projects",
                    self.project_id,
                    "badges",
                    mode="post",
                    **setting,
                    name=badge_name,
                )
            else:
                outdated = self._check_outdated(setting, resp_dict[badge_name])
                if outdated:
                    msg = "updated"
                    resp = self.api_request(
                        "projects",
                        self.project_id,
                        "badges",
                        resp_dict[badge_name]["id"],
                        mode="put",
                        **setting,
                    )
                else:
                    self._print(f"ok badge: {badge_name}", fg="green")
                    continue

            self._print(f"{msg} badge: {resp['name']}", fg="yellow")

    def _check_outdated(self, setting: tp.Any, resp: tp.Any) -> bool:
        if isinstance(setting, dict):
            for key, val in setting.items():
                val2 = self._special_key_resolution(resp, key)
                if self._check_outdated(val, val2):
                    self._print(f"{key}: {val}!={val2}", fg="magenta")
                    return True
        else:
            if str(setting) != str(resp):
                return True

        return False

    @classmethod
    def _special_key_resolution(cls, resp: tp.Any, key: str) -> tp.Any:
        if key in ["push_access_level", "merge_access_level"]:
            lvls = resp[f"{key}s"]
            assert len(lvls) == 1
            return lvls[0]["access_level"]
        return resp[key]

    def sync_protection(self) -> None:
        vip_branches = self.ns.auxe.gitlab.vip_branches
        resp = self.api_request("projects", self.project_id, "protected_branches")
        on_remote = [x["name"] for x in resp]
        to_remove = set(on_remote) - set(vip_branches)

        resp_dict = {x["name"]: x for x in resp}
        for branch, setting in vip_branches.items():
            if branch not in on_remote:
                msg = "installed"
            else:
                outdated = self._check_outdated(setting, resp_dict[branch])
                if outdated:
                    resp = self.api_request(
                        "projects",
                        self.project_id,
                        "protected_branches",
                        branch,
                        mode="delete",
                    )
                    msg = "updated"
                else:
                    self._print(f"ok protection: {branch}", fg="green")
                    continue

            resp = self.api_request(
                "projects",
                self.project_id,
                "protected_branches",
                **setting,
                name=branch,
                mode="post",
            )
            self._print(f"{msg} protection: {resp['name']}", fg="yellow")

        for branch in to_remove:
            resp = self.api_request(
                "projects", self.project_id, "protected_branches", branch, mode="delete"
            )
            self._print(f"removed protection for: {branch}", fg="red")

    def bake(self) -> None:
        self.sync_settings()
        self.sync_badges()
        self.sync_protection()

    @staticmethod
    def _col(x: tp.Any) -> tp.Union[str, tp.Tuple[int, int, int]]:
        status = x["status"]
        col: tp.Union[str, tp.Tuple[int, int, int]] = "yellow"
        if status == "success":
            col = "green"
        elif status == "failed":
            if x.get("allow_failure", False):
                col = (191, 255, 0)  # lime
            else:
                col = "red"
        elif status == "pending":
            col = "blue"
        elif status == "created":
            col = "cyan"
        elif status == "canceled":
            col = "magenta"
        return col

    def pipeline(self, show_success: bool = False) -> tp.Tuple[bool, bool]:
        out = subprocess.run(
            ["git", "rev-parse", "--abbrev-ref", "HEAD"],
            check=True,
            capture_output=True,
        )
        branch = out.stdout.decode().strip()
        resp = self.api_request("projects", self.project_id, "pipelines")
        for pipeline in resp:
            self.pretty_pipeline(pipeline)
            if branch not in pipeline["ref"]:
                continue
            self.display_pipeline(pipeline, show_success)

            done = bool(pipeline["status"] in ["success", "failed"])
            success = bool(pipeline["status"] in ["success"])
            return done, success
        return True, False

    def pretty_pipeline(self, pipeline: tp.Any) -> None:
        if "refs/merge-requests" in pipeline["ref"]:
            mr_id = int(pipeline["ref"].split("/")[2])
            mr_resp = self.api_request(
                "projects", self.project_id, "merge_requests", mr_id
            )
            pipeline["ref"] = "{source_branch}->{target_branch}".format_map(mr_resp)
        ddt = datetime.datetime
        start = ddt.strptime(pipeline["created_at"], "%Y-%m-%dT%H:%M:%S.%fZ")
        if pipeline["status"] != "running":
            last_update = ddt.strptime(pipeline["updated_at"], "%Y-%m-%dT%H:%M:%S.%fZ")
        else:
            last_update = datetime.datetime.utcnow()
        delta = last_update - start
        pipeline["duration"] = str(delta).rsplit(".", 1)[0].split(":", 1)[1]
        pipeline["source"] = pipeline["source"][:2]

    def display_pipeline(self, pipeline: tp.Any, show_success: bool = False) -> None:
        self._print(
            "[{source}|{status[0]}|{duration}]{ref:<36} {web_url}".format_map(pipeline),
            fg=self._col(pipeline),
        )
        jobs = self.api_request(
            "projects", self.project_id, "pipelines", pipeline["id"], "jobs"
        )

        for job in reversed(jobs):
            if not show_success and job["status"] == "success":
                continue
            job["name"] = "{stage}:{name}".format_map(job)
            if job["duration"] is None:
                job["duration"] = "--:--"
            else:
                job["duration"] = (
                    str(datetime.timedelta(seconds=job["duration"]))
                    .split(".", 1)[0]
                    .split(":", 1)[1]
                )
            self._print(
                "   [{status[0]}|{duration}]{name:<36} {web_url}".format_map(job),
                fg=self._col(job),
            )

    def merge_request(self, mr_message: str = "") -> None:
        data = self.ns.auxe.gitlab
        source_branch = data.current_branch
        target_branch = data.default_branch

        if source_branch == target_branch:
            raise RuntimeError(
                f"source and target branch are identical: {source_branch}"
            )
        coord = ["projects", self.project_id, "merge_requests"]

        settings = dict(
            source_branch=source_branch,
            target_branch=target_branch,
            remove_source_branch=True,
            squash=data.squash,
        )

        resp = self.api_request(*coord)
        for mr, _ in self._get_valid_mr(
            resp, mr_message=None, rerelease=True, settings=settings
        ):  # pylint: disable=invalid-name
            self._print(f"existing mr {mr['web_url']}", fg="yellow")
            break
        else:
            if not mr_message:
                self._print("not creating mr due to empty message", fg="red")
                return
            logger.info("creating mr %s->%s", source_branch, target_branch)
            mr = self.api_request(  # pylint: disable=invalid-name
                *coord,
                mode="post",
                **settings,
                title=mr_message,
            )

            self._print(f"created mr {mr['web_url']}", fg="green")

    def release(  # pylint: disable=too-many-locals
        self, rerelease: bool = False
    ) -> tp.Tuple[bool, bool]:
        current_version, _ = self.ns.get_current_version_and_lines()
        mr_message = f"release {current_version}"
        coord = ["projects", self.project_id, "merge_requests"]

        settings = dict(source_branch="develop", target_branch="release")
        release_notes = self.ns.get_release_notes()

        self._print(f"{mr_message} [{release_notes[current_version]}]", fg="green")
        resp = self.api_request(*coord)
        show = [  # order is important
            settings["target_branch"],
            "{source_branch}->{target_branch}".format_map(settings),
        ]
        n_success = len(show)
        error = 0
        success = 0

        for mr, needs_mwps in self._get_valid_mr(
            resp, mr_message, rerelease, settings
        ):  # pylint: disable=invalid-name
            if needs_mwps:
                resp = self.api_request(
                    *coord,
                    mr["iid"],
                    "merge",
                    mode="put",
                    merge_when_pipeline_succeeds=True,
                )
            resp = self.api_request("projects", self.project_id, "pipelines")
            for pipeline in resp:
                self.pretty_pipeline(pipeline)
                if pipeline["ref"] in show:
                    i = show.index(pipeline["ref"])
                    show = show[i + 1 :]  # we dont want to see the old tag pipeline
                    if pipeline["status"] != "success":
                        error += 1
                        self.display_pipeline(pipeline)
                    else:
                        success += 1

            if success == n_success and not error:
                self._print("fully released, congrats :)", fg="green")
                return True, True
            if not error:
                self._print(f"release ongoing {success}/{n_success}", fg="yellow")
            return False, False

        # create new mr
        mr = self.api_request(  # pylint: disable=invalid-name
            *coord,
            mode="post",
            **settings,
            title=mr_message,
        )
        for _ in range(5):
            if mr["merge_status"] == "can_be_merged":
                resp = self.api_request(
                    *coord,
                    mr["iid"],
                    "merge",
                    mode="put",
                    merge_when_pipeline_succeeds=True,
                )
                mwps_msg = " and set mwps"
                break
            time.sleep(1)
            mr = self.api_request(*coord, mr["iid"])  # pylint: disable=invalid-name
        else:
            mwps_msg = " and was NOT able to set mwps"

        self._print(
            "created release-mr ({source_branch}->{target_branch})".format_map(settings)
            + mwps_msg,
            fg="yellow",
        )
        return False, False

    def _get_valid_mr(
        self,
        mrs: tp.List[tp.Any],
        mr_message: tp.Optional[str],
        rerelease: bool,
        settings: tp.Dict[str, str],
    ) -> tp.Iterator[tp.Any]:
        for mr in mrs:
            # check that there isnt an (open) mr to release already
            if (
                mr["source_branch"] == settings["source_branch"]
                and mr["target_branch"] == settings["target_branch"]
            ):
                if mr_message and mr["title"] != mr_message:
                    continue

                if rerelease and mr["state"] != "opened":
                    # ignore closed MR with same name on rerelease
                    continue

                # if the pipeline fails, we set the merge again
                needs_mwps = (
                    not mr["merge_when_pipeline_succeeds"]
                    and mr["state"] == "opened"
                    and mr["merge_status"] == "can_be_merged"
                )
                yield mr, needs_mwps



# Copyright (c) 2021-2023 Mario S. Könz; License: MIT
import dataclasses as dc
import typing as tp

from ..._logging import logger
from ..._proto_namespace import _ProtoNamespace
from ._base import Payload

__all__ = ["DockerPayload"]


@dc.dataclass
class DockerPayload(Payload):
    param: _ProtoNamespace = dc.field(default_factory=_ProtoNamespace)

    def _get_executor(self, dep: Payload) -> tp.Any:
        if dep.flavor in ["docker_build", "with_dependency"]:
            return dep.executor  # type: ignore
        logger.info(
            "%s only extracts parents from type DockerBuildPayload and WithDependencyPayload[DockerBuildPayload]! No parent for %s",
            self.__class__.__name__,
            dep.flavor,
        )
        return None

    def create_executor(self, parents: tp.Any) -> tp.Any:
        raise NotImplementedError()

    def hydrate(self, deps: tp.Tuple[Payload, ...] = tuple()) -> None:
        if hasattr(self, "is_hydrated"):
            return
        # pylint: disable=attribute-defined-outside-init
        self.is_hydrated = True
        parents = [self._get_executor(dep) for dep in deps]
        parents = [par for par in parents if par is not None]
        if hasattr(self, "_executor"):
            assert self.executor.parents == parents
            return
        self._executor = self.create_executor(parents)

    @property
    def executor(self) -> tp.Any:
        return self._executor

    def run(self, force: bool = False) -> None:
        with self.write_environment_variables(self.param):
            if self.is_up_to_date():
                if force:
                    self.executor.script(force=force)
            self.executor.script()



# Copyright (c) 2021-2023 Mario S. Könz; License: MIT
import datetime
import os

from .._proto_namespace import _ProtoNamespace
from ._03_meta import MetaMixin


class ProjectMixin(MetaMixin):
    def templated(self, negative_default: bool = False) -> None:
        super().templated(negative_default)
        data = self.auxcon.setdefault("project", _ProtoNamespace())
        data.creation_year = self.get_current_year()

    def demodata(self) -> None:
        super().demodata()
        data = self.auxcon.project
        data.name = "hallo.world"
        data.slug = "hlw"
        data.author = "anonymous"
        data.license = "MIT"

    def formatted(self) -> None:
        super().formatted()
        keys = [
            "name",
            "slug",
            "author",
            "creation_year",
            "license",
            "source_dir",
        ]
        self._copy_keys_over(keys, "project")

    def defaulted(self) -> None:
        # pylint: disable=no-member
        super().defaulted()
        data = self.auxd.project
        data.setdefault("license", "Proprietary")
        data.setdefault("source_dir", "source")

    def enriched(self) -> None:
        super().enriched()
        data = self.auxe.project
        data.active_years = self.get_active_years()

    def bake(self) -> None:
        super().bake()
        # license
        data = self.auxe.project
        if data.license != "Proprietary":
            self.bake_file(
                f"license/{data.license}.txt",
                (self.target / ".." / "LICENSE.txt").resolve(),
            )

    @classmethod
    def deduce_user(cls) -> str:
        return os.environ.get("USER", os.environ.get("USERNAME", "unknown"))

    def get_active_years(self) -> str:
        current_year = self.get_current_year()
        creation_year = self.auxd.project.creation_year
        if creation_year == current_year:
            return f"{creation_year}"

        return f"{creation_year}-{current_year}"

    @classmethod
    def get_current_year(cls) -> str:
        return str(datetime.date.today().year)

    def step_migrate(self, to_version: str) -> None:
        super().step_migrate(to_version)

        if to_version == "3.0.0":
            data = self.auxcon.project
            pdata = _ProtoNamespace()
            for key in ["minimal_version", "supported_versions", "project_urls"]:
                if key in data:
                    pdata[key] = data.pop(key)
            if pdata:
                self.auxcon.python_project = pdata



# Copyright (c) 2021-2023 Mario S. Könz; License: MIT
import dataclasses as dc
import functools
import json
import os
import re
import subprocess
import sys
import typing as tp

import requests

from ._logging import logger
from ._proto_namespace import _ProtoNamespace
from ._yaml import yaml  # type: ignore

__all__ = ["LazyVersionStr", "EnvironmentVariables"]


class LazyVersionStr:
    def __str__(self) -> str:
        # pylint: disable=import-outside-toplevel,cyclic-import
        import adaux

        return adaux.__version__


@dc.dataclass
class ApiRequestCommunicator:
    base_url: str = dc.field(init=False)
    headers: tp.Dict[str, str] = dc.field(init=False)

    def url(self, *args: str) -> str:
        root = (self.base_url,)
        return "/".join(map(str, root + args))

    def api(self, *args: tp.Any) -> str:
        return self.url("api", "v4", *args)

    def api_request(self, *args: tp.Any, mode: str = "get", **kwgs: tp.Any) -> tp.Any:
        url = self.api(*args)
        if mode == "get":
            for key in kwgs:
                if key not in ["params"]:
                    raise RuntimeError("only params is allowed as kwgs")
            return self.get_request(url, **kwgs)
        if mode == "put":
            return self.put_request(url, json=kwgs)
        if mode == "post":
            return self.post_request(url, json=kwgs)
        if mode == "delete":
            return self.delete_request(url)

        raise NotImplementedError(mode)

    def graphql(self, *args: tp.Any) -> str:
        # https://gitlab.com/-/graphql-explorer
        return self.url("api", "graphql", *args)

    def graphql_request(
        self, mode: str = "get", **kwgs: tp.Any
    ) -> tp.Dict[str, tp.Any]:
        url = self.graphql()
        if mode == "get":
            return self.get_request(url, json=kwgs)
        raise NotImplementedError(mode)

    def get_request(self, url: str, **kwgs: tp.Any) -> tp.Dict[str, tp.Any]:
        req = requests.get(url, headers=self.headers, timeout=10, **kwgs)
        if req.status_code != 200:
            raise RuntimeError(f"{req} {req.text}")
        return json.loads(req.text)  # type: ignore

    def put_request(self, url: str, **kwgs: tp.Any) -> tp.Dict[str, tp.Any]:
        req = requests.put(url, headers=self.headers, timeout=10, **kwgs)
        self.assert_with_debug(req, 200)
        return json.loads(req.text)  # type: ignore

    def post_request(self, url: str, **kwgs: tp.Any) -> tp.Dict[str, tp.Any]:
        req = requests.post(url, headers=self.headers, timeout=10, **kwgs)
        self.assert_with_debug(req, 201)
        return json.loads(req.text)  # type: ignore

    def delete_request(self, url: str, **kwgs: tp.Any) -> tp.Dict[str, tp.Any]:
        req = requests.delete(url, headers=self.headers, timeout=10, **kwgs)
        self.assert_with_debug(req, 204)
        return {}

    @classmethod
    def assert_with_debug(cls, req: tp.Any, code: int) -> None:
        if req.status_code != code:
            print(req, req.text)
            assert req.status_code == code


def subprocess_run(
    *args: tp.Any, **kwgs: tp.Any
) -> "subprocess.CompletedProcess[bytes]":
    sys.stdout.flush()
    try:
        # pylint: disable=subprocess-run-check
        return subprocess.run(*args, **kwgs)
    except subprocess.CalledProcessError as err:
        if err.stdout:
            print("========>> stdout")
            print(err.stdout.decode("utf-8"))
        if err.stderr:
            print("========>> stderr")
            print(err.stderr.decode("utf-8"))
        raise


@dc.dataclass(frozen=True)
class EnvironmentVariables:
    aux: _ProtoNamespace
    regex_release: str = r"^(?:v)?(\d+)\.(\d+)\.(\d+)$"
    regex_version: str = r"(?:v)?(\d+)\.(\d+)\.(\d+)"

    def apply_format(self, val: str) -> str:
        if "{incr}" in val:
            val = val.replace("{incr}", "{{incr}}")
        res = val.format_map(self)
        if "{incr}" in val:
            res = self.auto_incr_process(res)
        return res

    @functools.cached_property
    def release_note(self) -> str:
        release_notes = self.aux.python_project.release_notes
        version = self.aux.python_project.version
        res: str = release_notes[version]
        return res

    @functools.cached_property
    def version(self) -> str:
        res: str = self.aux.python_project.version
        return res

    @functools.cached_property
    def branch(self) -> str:
        res: str = self.aux.gitlab.current_branch
        return res

    @functools.cached_property
    def ci_adaux_image(self) -> str:
        res: str = self.aux.versions.ci_adaux_image
        return res

    @functools.cached_property
    def commit_tag(self) -> str:
        res = os.environ.get("CI_COMMIT_TAG", "")
        if res:
            return res
        reason = "empty" if "CI_COMMIT_TAG" in os.environ else "inexistent"
        raise ValueError(f"commit_tag is {reason}")

    @functools.cached_property
    def _remote_version_tags(self) -> tp.List[str]:
        out = subprocess_run(
            ["git", "ls-remote", "-q", "--tags"], check=True, capture_output=True
        )

        res = list(re.findall(r"tags/(.+)\n", out.stdout.decode()))
        return self._version_sorted(res)

    def _version_sorted(self, tags: tp.List[str]) -> tp.List[str]:
        res = []
        for tag in tags:
            match = re.match(self.regex_version, tag)
            if match:
                major, minor, patch = match.groups()
                res.append((int(major), int(minor), int(patch), tag))
            else:
                res.append((0, 0, 0, tag))
        res = sorted(res)
        return [x[-1] for x in res]

    @functools.cached_property
    def last_release(self) -> str:
        for tag in reversed(self._remote_version_tags):
            match = re.match(self.regex_release, tag)
            if match:
                return tag
        return "0.0.0"

    def auto_incr_process(self, tag: str) -> str:
        pattern = tag.replace("{incr}", r"(\d+)")
        candidates = []
        for candidate in reversed(self._remote_version_tags):
            match = re.match(f"^{pattern}$", candidate)
            if match:
                (incr,) = match.groups()
                candidates.append((int(incr), candidate))

        if candidates:
            incr, candidate = list(sorted(candidates))[-1]
            incr_val = incr + 1
        else:
            incr_val = 1
        return tag.format(incr=incr_val)

    @functools.cached_property
    def autoincr(self) -> str:
        tags_used = self._remote_version_tags
        for tag in tags_used:
            logger.info("tag %s available", tag)
        for tag in reversed(tags_used):
            # get the first tag that matches the format
            regex = self.regex_version
            match = re.match(regex, tag)

            if match:
                major, minor, patch = match.groups()
                return f"{major}.{minor}.{int(patch)+1}"

        raise ValueError(f"no tag adheres to {regex}")

    def __getitem__(self, key: str) -> str:
        try:
            res: str = getattr(self, key)
        except AttributeError:
            # fallback on env
            if key in os.environ:
                res = os.environ[key]
            else:
                raise
        return res


def repr_env_var(
    representer: tp.Any, env_var: tp.Any  # pylint: disable=unused-argument
) -> tp.Any:
    return representer.represent_str("n/a")


yaml.representer.add_representer(EnvironmentVariables, repr_env_var)



# Copyright (c) 2021-2023 Mario S. Könz; License: MIT
import dataclasses as dc
import typing as tp

from ._base import Payload


__all__ = ["DockerComposePayload"]


@dc.dataclass
class DockerComposePayload(Payload):
    flavor: tp.ClassVar[str] = "docker_compose"
    payloads: tp.List[Payload]

    def run(self, force: bool = False) -> None:
        raise NotImplementedError()



# Copyright (c) 2021-2023 Mario S. Könz; License: MIT
from . import _base
from . import _docker
from . import _docker_build
from . import _docker_compose
from . import _docker_run
from . import _python
from . import _with_dependency
from ._base import *
from ._docker import *
from ._docker_build import *
from ._docker_compose import *
from ._docker_executors import subprocess_run
from ._docker_run import *
from ._python import *
from ._with_dependency import *

__all__ = (
    _base.__all__
    + _docker_build.__all__
    + _docker_run.__all__
    + _docker_compose.__all__
    + _python.__all__
    + _with_dependency.__all__
    + _docker.__all__
    + ["subprocess_run"]
)



# Copyright (c) 2021-2023 Mario S. Könz; License: MIT
import shlex
import typing as tp

import click

from ._cli import adaux
from ._cli import convert_runtime_to_click_error


@adaux.group()  # type: ignore[unused-ignore,misc]
@click.pass_context
def mm(ctx: click.Context) -> None:  # pylint: disable=unused-argument
    """
    Allows to run adaux or bash commands over multiple adaux projects.
    """


@mm.command()  # type: ignore[unused-ignore,misc]
@click.pass_context
@click.option(
    "-a",
    "--all-vip-branches",
    is_flag=True,
    default=False,
    help="update all vip branches",
)
@click.option(
    "-p",
    "--prune",
    is_flag=True,
    default=False,
    help="prunes local branches that are not on remote",
)
def update(
    ctx: click.Context, all_vip_branches: tp.Tuple[str, ...], prune: bool
) -> None:
    """
    Checkout and pull default branches.
    """
    with convert_runtime_to_click_error(ctx):
        with ctx.obj.extra():
            ctx.obj.mm_update(all_vip_branches, prune)


@mm.command()  # type: ignore[unused-ignore,misc]
@click.pass_context
@click.argument("cmd_parts", nargs=-1)
def run(ctx: click.Context, cmd_parts: tp.Tuple[str]) -> None:
    """
    Run bash command.
    """
    with convert_runtime_to_click_error(ctx):
        with ctx.obj.extra():
            if len(cmd_parts) == 0:
                tasks = ctx.obj.auxe.multimanage.tasks
                if len(tasks) == 0:
                    ctx.exit(1)
                click.secho("Available Tasks:", fg="yellow")
                for task in tasks:
                    print(f"- {task}")
                return
            if len(cmd_parts) > 1:
                cmd_str = shlex.join(cmd_parts)
            else:
                cmd_str = cmd_parts[0]
            ctx.obj.mm_run(cmd_str)


@mm.command()  # type: ignore[unused-ignore,misc]
@click.pass_context
@click.argument("cmd_parts", nargs=-1)
def a(ctx: click.Context, cmd_parts: tp.Tuple[str]) -> None:
    """
    Run adaux command / Shorthand for "adaux mm run adaux"
    """
    with convert_runtime_to_click_error(ctx):
        with ctx.obj.extra():
            if len(cmd_parts) == 0:
                ctx.exit(1)
            elif len(cmd_parts) > 1:
                cmd_str = shlex.join(cmd_parts)
            else:
                cmd_str = cmd_parts[0]
            ctx.obj.mm_adaux(cmd_str)



# Copyright (c) 2021-2023 Mario S. Könz; License: MIT
import dataclasses as dc
import functools
import hashlib
import logging
import os
import re
import shutil
import subprocess
import typing as tp
from pathlib import Path

import requests

from ..._logging import logger
from ..._util import subprocess_run


@dc.dataclass(frozen=True)
class DockerShared:
    slug: str
    parents: tp.Tuple["DockerShared", ...] = tuple()
    images: tp.List[str] = dc.field(default_factory=list)

    # Run local registry:
    # docker run -d -p 5000:5000 --name registry registry:2
    # docker container stop registry && docker container rm -v registry
    # export CI_REGISTRY=localhost:5000
    @property
    def registry_host(self) -> str:
        if not self.remote_exists():
            raise RuntimeError(
                "Your run needs a remote registry.\n"
                "In case you have not already, you can launch a local registry with\n"
                "  docker run -d -p 5000:5000 --name registry registry:2\n"
                "and run\n"
                "  export CI_REGISTRY=localhost:5000"
            )
        return os.environ["CI_REGISTRY"]

    def remote_exists(self) -> bool:
        return "CI_REGISTRY" in os.environ

    @property
    def build_deps(self) -> tp.Iterator["DockerBuildMixin"]:
        for parent in self.parents:
            if isinstance(parent, DockerBuildMixin):
                yield parent

    def script(self, force: bool = False) -> None:  # pylint: disable=unused-argument
        # get dependencies
        for job in self.build_deps:
            local_tag = job.local_tag
            # if not exists_locally(local_tag):
            #     if self.remote_exists():
            #         remote_tag = job.remote_tag
            #         if not exists_locally(remote_tag):
            #             cmd = ["docker", "pull", remote_tag]
            #             subprocess_run(cmd, check=True, capture_output=True)
            #             logger.info("pulled '%s'", remote_tag)

            #         tag_image(remote_tag, job.local_tag)
            if self.remote_exists() and local_tag:
                remote_tag = job.remote_tag
                cmd = ["docker", "pull", remote_tag]
                subprocess_run(cmd, check=True, capture_output=True)
                logger.info("pulled '%s'", remote_tag)
                tag_image(remote_tag, job.local_tag)

            # I tag every time as it allows different dependencies to
            # map to the same service name without raising issues
            # (useful for local development)
            tag_image(job.local_tag, job.use_tag)

        # download other images which are not a direct dependency (cross-repo)
        for image_tag in self.images:
            if self.remote_exists():
                image_remote_tag = (
                    f"{self.registry_host}/{self._remove_namespace_local(image_tag)}"
                )
                check_and_pull_if_not_existent(image_remote_tag, image_tag)
            else:
                tag = self._remove_namespace_local(image_tag)
                tag_image(tag, image_tag)

    def _get_tag_parts(self) -> tp.Iterator[tp.Union[bytes, str]]:
        for job in self.build_deps:
            yield job.tag

    def _remove_namespace_local(self, image: str) -> str:
        if not self.remote_exists() or self.registry_host.startswith("localhost"):
            return image.rsplit("/", 1)[1]
        return image


@dc.dataclass(frozen=True)
class _DockerBuildMixinState:
    service: str
    image_name: str


@dc.dataclass(frozen=True)
class DockerBuildMixin(DockerShared, _DockerBuildMixinState):
    files: tp.List[str] = dc.field(default_factory=list)
    always_build: bool = False

    def script(self, force: bool = False) -> None:
        # this tag need to be evaluated before super script, otherwise base-matching hasnt happened.
        local_tag = self.local_tag  # triggeres matching
        super().script(force=force)
        if self.remote_exists():
            remote_tag = self.remote_tag
            check_and_pull_if_not_existent(remote_tag, local_tag, self.service)

        if not exists_locally(self.local_tag) or self.always_build or force:
            cmd = self._get_build_cmd()
            logger.info(" ".join(cmd))
            subprocess_run(cmd, check=True)

        # upload
        tag_image(self.use_tag, local_tag)
        if self.remote_exists():
            upload_to_remote(local_tag, remote_tag)
            logger.info("uploaded %s to %s", local_tag, remote_tag)

    def _get_build_cmd(self) -> tp.List[str]:
        cmd = [
            "docker",
            "--log-level",
            "ERROR",
            "compose",
            "-p",
            self.slug,
            "-f",
            "devops/docker/compose.yml",
            "build",
            self.service,
        ]
        if not self.remote_exists():
            cmd.append("--pull=false")
        return cmd

    @functools.cached_property
    def tag(self) -> str:
        hasher = hashlib.md5()

        for part in self._get_tag_parts():
            if isinstance(part, str):
                hasher.update(part.encode("utf8"))
            else:
                hasher.update(part)

        tag = hasher.hexdigest()[:16]
        if self.remote_exists():
            return f"aux-{tag}"
        return f"loc-{tag}"

    def _get_tag_parts(self) -> tp.Iterator[tp.Union[bytes, str]]:
        yield from super()._get_tag_parts()
        for file in self.files:
            with open(file, "rb") as f:
                yield f.read()

        # we assume the image tag is not moving (i.e. no latest)
        yield from self.images

    @functools.cached_property
    def remote_tag(self) -> str:
        return f"{self.registry_host}/{self.local_tag}"

    @functools.cached_property
    def local_tag(self) -> str:
        image = self.get_image_name()
        return f"{image}:{self.tag}"

    @functools.cached_property
    def use_tag(self) -> str:
        return self.image_name.lower()

    def get_image_name(self) -> str:
        if "CI_PROJECT_NAMESPACE" in os.environ:
            group = os.environ["CI_PROJECT_NAMESPACE"]
            repo = os.environ["CI_PROJECT_NAME"]
            return f"{group}/{repo}/{self.use_tag}".lower()
        return f"{self.use_tag}"

    def _docker_registry_header_and_url(
        self, image: str
    ) -> tp.Tuple[tp.Dict[str, str], str]:
        protocol = "http"
        headers = {}
        if not self.registry_host.startswith("localhost"):
            protocol = "https"
            headers["Authorization"] = f"Bearer {registry_api_token(image)}"

        # https://gitlab.com/gitlab-org/gitlab/-/issues/19470
        url = f"{protocol}://{self.registry_host}/v2"
        return headers, url

    def is_up_to_date(self) -> bool:
        if self.remote_exists():
            return self.exists_remotely()
        return exists_locally(self.local_tag)

    def exists_remotely(self) -> bool:
        headers, api_url = self._docker_registry_header_and_url(self.get_image_name())
        try:
            req = requests.get(
                api_url + f"/{self.get_image_name()}/tags/list",
                headers=headers,
                timeout=10,
            )
        except requests.ConnectionError as err:
            raise RuntimeError(
                f"are you sure the registry {self.registry_host} is reachable?\n"
                "use: docker run -d -p 5000:5000 --name registry registry:2\n"
                "or: unset CI_REGISTRY"
            ) from err

        if req.status_code != 200:
            return False
        tags = req.json()["tags"] or []
        return self.tag in tags

    def tag_and_upload(self, tag: str) -> tp.Tuple[str, str]:
        local_tag = self.local_tag
        if self.remote_exists():
            remote_tag = self.remote_tag
            assert isinstance(self, DockerBuild)
            check_and_pull_if_not_existent(remote_tag, local_tag)
            release_tag = remote_tag.replace(self.tag, tag)
            upload_to_remote(local_tag, release_tag)
        else:
            logger.warning(
                "no remote exists (export CI_REGISTRY=localhost:5000), fallback on local tagging"
            )
            release_tag = local_tag.replace(self.tag, tag)
            tag_image(local_tag, release_tag)
        return local_tag, release_tag

    def pull_if_not_existent(self) -> str:
        remote_tag = self.remote_tag
        local_tag = self.local_tag
        assert isinstance(self, DockerBuild)
        check_and_pull_if_not_existent(remote_tag, local_tag)
        return local_tag


@dc.dataclass(frozen=True)
class _DockerBuildBranchMatchMixinState(_DockerBuildMixinState):
    branch_match: tp.List[tp.Tuple[str, str, str, tp.List[str], tp.Optional[str]]]


@dc.dataclass(frozen=True)
class MatchCommonImpl:
    @classmethod
    def _get_git_branch(cls) -> str:
        branch = os.environ.get("CI_COMMIT_BRANCH", "") + os.environ.get(
            "CI_MERGE_REQUEST_SOURCE_BRANCH_NAME", ""
        )
        if not branch:
            # we are running it locally, and try to get it from git
            proc = subprocess.run(
                ["git", "rev-parse", "--abbrev-ref", "HEAD"],
                check=False,
                capture_output=True,
            )
            if proc.returncode == 0:
                branch = proc.stdout.decode("utf-8").strip()

        assert branch
        return branch


@dc.dataclass(frozen=True)
class DockerBuildBranchMatchMixin(
    MatchCommonImpl, DockerBuildMixin, _DockerBuildBranchMatchMixinState
):
    def _get_tag_parts(self) -> tp.Iterator[tp.Union[bytes, str]]:
        yield from super()._get_tag_parts()

        branch = self._get_git_branch()

        for pkg, var, url, skip, fallback in self.branch_match:
            if branch in skip:
                logger.info("skipping matching branch '%s' for '%s'", branch, pkg)
                continue
            secret = re.search(r"\$(\w+)@", url).group(1)  # type: ignore
            url = url.replace(f"${secret}", os.environ[secret])
            # figure out if matching branch exists

            try_branches = [(branch, "matching")]
            if fallback:
                try_branches.append((fallback, "fallback"))

            for branch, reason in try_branches:
                exists, sha = self._matching_branch_exist(url, branch)
                if exists:
                    logger.info("found %s branch '%s' for '%s'", reason, branch, pkg)
                    yield sha
                    os.environ[var] = f"{pkg}@git+{url}@{branch}"
                    break

    @classmethod
    def _matching_branch_exist(cls, url: str, branch: str) -> tp.Tuple[bool, str]:
        cmd = f"git ls-remote {url} {branch}".split(" ")
        res = subprocess_run(cmd, check=False, capture_output=True)
        if res.returncode != 0:
            raise RuntimeError(res.stderr.decode("utf8"))
        parts = res.stdout.decode("utf8").split()
        if parts:
            sha, refs = parts
            assert branch in refs
            return True, sha
        return False, ""


@dc.dataclass(frozen=True)
class _DockerBuildBaseMatchMixinState(_DockerBuildMixinState):
    base_match: tp.List[tp.Tuple[str, str, tp.List[str], str, str]]


@dc.dataclass(frozen=True)
class DockerBuildBaseMatchMixin(
    MatchCommonImpl, DockerBuildMixin, _DockerBuildBaseMatchMixinState
):
    def _get_tag_parts(self) -> tp.Iterator[tp.Union[bytes, str]]:
        yield from super()._get_tag_parts()
        branch = self._get_git_branch()

        for image_w_ns, var, skip, fallback, default in self.base_match:
            # locally, we do not use the group and repo prefix, hence
            # it needs to be removed
            image = self._remove_namespace_local(image_w_ns)
            if branch in skip:
                logger.info(
                    "skipping matching base '%s' for '%s', using default '%s'",
                    branch,
                    image,
                    default,
                )
                self._fix_environ_and_image(image, var, default)
                yield default
                return

            try_bases = [(fallback, "fallback")]

            for tag, reason in try_bases:
                if not self.remote_exists():
                    self._call_upstream_adaux(image_w_ns, image, tag)

                exists, sha = self._matching_base_exist(image, tag)
                if exists:
                    logger.info("found %s tag '%s' for '%s'", reason, tag, image)
                    self._fix_environ_and_image(image, var, tag)
                    yield sha
                    break
            else:
                logger.info(
                    "no matching base found for '%s', using default '%s'",
                    image,
                    default,
                )
                self._fix_environ_and_image(image, var, default)
                yield tag

    def _fix_environ_and_image(self, image: str, var: str, tag: str) -> None:
        unevaluated = f"{image}:${var}"
        os.environ[var] = tag

        for i, candidate in enumerate(self.images):
            if candidate.endswith(unevaluated):
                self.images[i] = candidate.replace(f"${var}", tag)

    def _matching_base_exist(self, image: str, tag: str) -> tp.Tuple[bool, str]:
        if self.remote_exists():
            headers, api_url = self._docker_registry_header_and_url(image)
            try:
                headers[
                    "Accept"
                ] = "application/vnd.docker.distribution.manifest.v2+json"
                req = requests.get(
                    api_url + f"/{image}/manifests/{tag}", headers=headers, timeout=10
                )
                sha = req.json()["config"]["digest"]
                logger.info("base matched image %s:%s has digest %s", image, tag, sha)
                return True, sha
            except (requests.ConnectionError, KeyError):
                return False, ""
        else:
            res = subprocess_run(
                "docker inspect --format='{{.Id}}'".split(" ") + [f"{image}:{tag}"],
                capture_output=True,
            )
            if res.returncode == 0:
                return True, res.stdout.decode("utf-8")
            return False, ""

    def _call_upstream_adaux(self, image_w_ns: str, image: str, tag: str) -> None:
        _, project, image = image_w_ns.split("/")
        orig_path = Path.cwd()
        slug = image.split("-", 1)[0]
        candidates = [project]
        if "_" in project:
            candidates += [project.replace("_", "-"), project.rsplit("_", 1)[1]]
        if "-" in project:
            candidates += [project.replace("-", "_"), project.rsplit("-", 1)[1]]
        for cand in candidates:
            upstream = orig_path.parent / cand
            if upstream.exists():
                break
        else:
            logger.warning(
                "project %s not found as a sibling directory, checked for %s",
                project,
                candidates,
            )
            return

        verbose = ""
        if logger.getEffectiveLevel() == logging.DEBUG:
            verbose = " -v"
        os.chdir(upstream)
        with subprocess.Popen(
            f"adaux{verbose} run tag-{tag}".split(" "),
            stdout=subprocess.PIPE,
            text=True,
        ) as process:
            assert process.stdout is not None
            columns, _ = shutil.get_terminal_size((80, 20))
            print(">" * columns)
            for line in iter(process.stdout.readline, ""):
                prefix = f"|{slug}|"
                if line.startswith("|"):
                    if verbose:
                        prefix = "|"
                    else:
                        prefix = ""
                line = prefix + line
                # adjust column wide lines as terminal size is wrong in subcall
                if ">>>>>>>" in line:
                    line = line.split(">", 1)[0]
                    line += ">" * (columns - len(line)) + "\n"
                if "<<<<<<<" in line:
                    line = line.split("<", 1)[0]
                    line += "<" * (columns - len(line)) + "\n"
                print(f"{line}", end="", flush=True)
            print("<" * columns, flush=True)
        os.chdir(orig_path)


@dc.dataclass(frozen=True)
class _DockerRunMixinState:
    services: tp.List[str]


@dc.dataclass(frozen=True)
class DockerRunMixin(DockerShared, _DockerRunMixinState):
    interactive: bool = False
    exec_join: bool = False
    run_ports: tp.List[str] = dc.field(default_factory=list)

    def script(self, force: bool = False) -> None:
        super().script(force=force)
        base_cmd = [
            "docker",
            "--log-level",
            "ERROR",
            "compose",
            "-p",
            self.slug,
            "-f",
            "devops/docker/compose.yml",
        ]
        if self.interactive:
            if len(self.services) > 1:
                raise NotImplementedError(
                    "cannot run more than one service in interactive mode"
                )
            service = self.services[0]
            sub_cmd = ["run", "--rm", "-ti"]
            for port in self.run_ports:
                sub_cmd.append("-p")
                sub_cmd.append(port)

            if self.exec_join:
                res = subprocess_run(
                    [
                        "docker",
                        "ps",
                        "--filter",
                        f"label=com.docker.compose.service={service}",
                        "--format",
                        "{{.ID}}",
                    ],
                    capture_output=True,
                )
                container_id = res.stdout.decode("utf-8").strip()
                logger.info(
                    "using exec instead of run, as container %s=%s exists",
                    service,
                    container_id,
                )
                if container_id:
                    res = subprocess_run(
                        [
                            "docker",
                            "inspect",
                            "--format",
                            "{{.Config.Cmd}}",
                            container_id,
                        ],
                        capture_output=True,
                    )
                    # very simple listification for now
                    orig_cmd = res.stdout.decode("utf-8").strip()[1:-1].split(",")
                    # service instead of container_id bc of compose
                    sub_cmd = ["exec", "-ti", service] + orig_cmd
                else:
                    sub_cmd.append(service)
            else:
                sub_cmd.append(service)
            cmd = base_cmd + sub_cmd
        else:
            cmd = base_cmd + [
                "up",
                "--no-build",
                "--abort-on-container-exit",
            ]
            if len(self.services) <= 1:
                cmd.append("--no-log-prefix")
            cmd += self.services

        subprocess_run(cmd, check=True)


@dc.dataclass(frozen=True)
class DockerBuild(
    DockerBuildBaseMatchMixin,
    DockerBuildBranchMatchMixin,
    _DockerBuildBaseMatchMixinState,
    _DockerBuildBranchMatchMixinState,
):
    pass


@dc.dataclass(frozen=True)
class DockerBuildx(DockerBuild):
    platforms: tp.List[str] = dc.field(default_factory=list)

    def _get_build_cmd(self) -> tp.List[str]:
        cmd = super()._get_build_cmd()
        idx = cmd.index("-p")
        cmd.remove("-p")
        cmd.pop(idx)

        idx = cmd.index("compose")
        cmd[idx] = "buildx"
        cmd.insert(idx + 1, "bake")
        cmd.remove("build")
        cmd.append("--provenance=0")
        if self.remote_exists():
            cmd.append("--push")
        else:
            cmd.append("--set=*.platform=linux/arm64")
            cmd.append("--load")
        return cmd


@dc.dataclass(frozen=True)
class DockerRun(DockerRunMixin, _DockerRunMixinState):
    pass


# try to download for caching (or reuse if already there)
def check_and_pull_if_not_existent(
    remote_tag: str, local_tag: str, service: tp.Optional[str] = None
) -> None:
    cmd = ["docker", "pull", remote_tag]
    res = subprocess_run(cmd, check=False, capture_output=True)
    if res.returncode != 0:
        return

    tag_image(remote_tag, local_tag)
    if service:
        tag_image(remote_tag, service)


def upload_to_remote(
    local_tag: str, remote_tag: str, registry: tp.Optional[str] = None
) -> None:
    tag_image(local_tag, remote_tag)
    cmd = ["docker", "push", remote_tag]
    if registry:
        cmd += [registry]
    subprocess_run(cmd, check=True, capture_output=True)
    logger.info("pushed '%s'", remote_tag)


def tag_image(src_tag: str, dest_tag: str) -> None:
    cmd = ["docker", "image", "tag", src_tag, dest_tag]
    subprocess_run(cmd, check=True, capture_output=True)
    logger.info("tagged '%s' -> '%s'", src_tag, dest_tag)


def exists_locally(local_tag: str) -> bool:
    cmd = ["docker", "image", "inspect", local_tag]
    res = subprocess_run(cmd, check=False, capture_output=True)
    return res.returncode == 0


@functools.lru_cache
def registry_api_token(image: str) -> str:
    # https://www.pimwiddershoven.nl/entry/request-an-api-bearer-token-from-gitlab-jwt-authentication-to-control-your-private-docker-registry
    # curl --user 'user:pw' 'https://gitlab.com/jwt/auth?client_id=docker&offline_token=true&service=container_registry&scope=repository:administratum/auxilium/adaux-pytest-39:pull,push'
    # curl -H 'Authorization: Bearer token' https://gitlab.com:5050/v2/administratum/auxilium/adaux-pytest-39/tags/list
    server = os.environ["CI_SERVER_URL"]
    url = f"{server}/jwt/auth?client_id=docker&offline_token=true&service=container_registry&scope=repository:{image}:pull,push"
    req = requests.get(
        url,
        auth=(os.environ["CI_REGISTRY_USER"], os.environ["CI_JOB_TOKEN"]),
        timeout=10,
    )
    return req.json()["token"]  # type: ignore



# Copyright (c) 2021-2023 Mario S. Könz; License: MIT
import collections
import contextlib
import typing as tp

from .._logging import logger
from .._proto_namespace import _ProtoNamespace
from ._00_extra_level import ExtraLevel
from ._03_meta import MetaMixin


class SentinelMixin(MetaMixin):
    def __init__(self, *args: tp.Any, **kwgs: tp.Any) -> None:
        super().__init__(*args, **kwgs)
        self._auxcon_entered = False
        self._extra_counter: tp.DefaultDict[ExtraLevel, int] = collections.defaultdict(
            int
        )

    def _part_enabled(self, part: tp.Type[tp.Any]) -> bool:  # MetaMixin related
        enable, _ = self._en_disabled_list()
        if enable:
            mro = self.__class__.__mro__
            if SentinelMixin in mro:  # keep after (and)
                if mro.index(part) <= mro.index(SentinelMixin):
                    return True

        return super()._part_enabled(part)

    def bake(self) -> None:
        with self.extra():
            super().bake()
            self.writeout()

    @contextlib.contextmanager
    def with_formatted(self) -> tp.Iterator[_ProtoNamespace]:
        level = ExtraLevel.FORMATTED
        if self._extra_counter[level] == 0:
            super().formatted()
        self._extra_counter[level] += 1
        yield self.auxf
        self._extra_counter[level] -= 1
        if self._extra_counter[level] == 0:
            del self.auxf

    @contextlib.contextmanager
    def with_defaulted(self) -> tp.Iterator[_ProtoNamespace]:
        level = ExtraLevel.DEFAULTED
        if self._extra_counter[level] == 0:
            super().defaulted()
        self._extra_counter[level] += 1
        yield self.auxd
        self._extra_counter[level] -= 1
        if self._extra_counter[level] == 0:
            del self.auxd

    @contextlib.contextmanager
    def with_enriched(self) -> tp.Iterator[_ProtoNamespace]:
        level = ExtraLevel.ENRICHED
        if self._extra_counter[level] == 0:
            super().enriched()
        self._extra_counter[level] += 1
        yield self.auxe
        self._extra_counter[level] -= 1
        if self._extra_counter[level] == 0:
            del self.auxe

    @contextlib.contextmanager
    def with_hydrated(self) -> tp.Iterator[_ProtoNamespace]:
        level = ExtraLevel.HYDRATED
        if self._extra_counter[level] == 0:
            super().hydrated()
        self._extra_counter[level] += 1
        yield self.auxh
        self._extra_counter[level] -= 1
        if self._extra_counter[level] == 0:
            del self.auxh

    @contextlib.contextmanager
    def extra(
        self, level: ExtraLevel = ExtraLevel.ENRICHED
    ) -> tp.Iterator[_ProtoNamespace]:
        if self._extra_counter[level] == 0:
            logger.debug("enabled %s", level)
        if level == ExtraLevel.TEMPLATED:
            self.templated()
            with self.with_formatted() as aux:
                yield aux
        elif level == ExtraLevel.TEMPLATED_WITH_NEGATIVE:
            self.templated(negative_default=True)
            with self.with_formatted() as aux:
                yield aux
        elif level == ExtraLevel.DEMODATA:
            self.templated()
            self.demodata()
            with self.with_formatted() as aux:
                yield aux
        if level == ExtraLevel.RAW:
            yield self.auxcon
        elif level == ExtraLevel.FORMATTED:
            with self.with_formatted() as aux:
                yield aux
        elif level == ExtraLevel.DEFAULTED:
            with self.with_formatted(), self.with_defaulted() as aux:
                yield aux
        elif level == ExtraLevel.ENRICHED:
            with self.with_formatted(), self.with_defaulted(), self.with_enriched() as aux:
                yield aux
        elif level == ExtraLevel.HYDRATED:
            with self.with_formatted(), self.with_defaulted(), self.with_enriched(), self.with_hydrated() as aux:
                yield aux

        if self._extra_counter[level] == 0:
            logger.debug("disabled %s", level)



# Copyright (c) 2021-2023 Mario S. Könz; License: MIT
import sys
import typing as tp
from pathlib import Path

from .._logging import logger
from .._parser import ConfigParser
from .._parser import Jinja2Parser
from .._proto_namespace import _ProtoNamespace
from ._05_project import ProjectMixin


class PythonProjectMixin(ProjectMixin):
    def templated(self, negative_default: bool = False) -> None:
        super().templated(negative_default)
        self.auxcon.setdefault("python_project", _ProtoNamespace())

    def demodata(self) -> None:
        super().demodata()
        data = self.auxcon.python_project
        data.minimal_version = (self.deduce_python_version(),)

    def formatted(self) -> None:
        super().formatted()
        keys = [
            "minimal_version",
            "supported_versions",
            "project_urls",
        ]
        self._copy_keys_over(keys, "python_project")
        self._to_list("python_project", "project_urls")

    def defaulted(self) -> None:
        # pylint: disable=no-member
        super().defaulted()
        data = self.auxd.python_project
        defaults = dict(
            project_urls=_ProtoNamespace(),
            supported_versions=[self.auxf.python_project.minimal_version],
        )
        for key, val in defaults.items():
            data.setdefault(key, val)

    def enriched(self) -> None:
        super().enriched()
        pro = self.auxe.project
        data = self.auxe.python_project
        data.minimal_version_slug = data.minimal_version.replace(".", "")
        data.module_dir = pro.source_dir + "/" + pro.name.replace(".", "/")
        data.namespace_name, data.second_name = "", pro.name
        if "." in pro.name:
            data.namespace_name, data.second_name = pro.name.split(".", 1)
        data.setup_fields = [
            "name",
            "author",
            "license",
            "description",
            "long_description",
            "project_urls",
        ]
        data.version, _ = self.get_current_version_and_lines(data)
        data.release_notes = self.get_release_notes(self.auxd)

    def bake(self) -> None:
        super().bake()
        pro = self.auxe.project
        data = self.auxe.python_project
        self.bake_file("install-dev.sh", chmod=0o755)
        self.bake_file("root/_setup.py", "../setup.py")

        srcj = self.root / "root/setup.cfg.jinja2"
        with Jinja2Parser.render_to_tmp(srcj, aux=self.auxe) as src:
            data.config = ConfigParser.read(src)
            for key in data.setup_fields:
                if key in data:
                    val = data[key]
                elif key in pro:
                    val = pro[key]
                else:
                    continue

                if val:
                    data.config.metadata[key] = val

    def writeout(self) -> None:
        super().writeout()
        if self.is_enabled("pip"):
            dest = self.target / "../setup.cfg"
            written = ConfigParser.write(self.auxe.python_project.pop("config"), dest)
            if written:
                self._print(f"baked {dest}", fg="green")

    @classmethod
    def deduce_project_name(cls, path: tp.Optional[Path] = None) -> str:
        path = path or (Path.cwd())

        # level 1
        for obj in path.glob("*/__init__.py"):
            with obj.open("r", encoding="utf-8") as f:
                if "__version__" in f.read():
                    lvl1 = obj.parent.stem
                    return lvl1
        # level 2
        for obj in path.glob("*/*/__init__.py"):
            with obj.open("r", encoding="utf-8") as f:
                if "__version__" in f.read():
                    lvl1 = obj.parent.stem
                    lvl2 = obj.parent.parent.stem
                    if lvl2 in ["source"]:
                        return lvl1
                    return f"{lvl2}.{lvl1}"
        return "not-found"

    @classmethod
    def deduce_project_slug(cls) -> str:
        proj_name = cls.deduce_project_name()
        if proj_name.count(".") == 1:
            ns, sub = proj_name.split(".")
            return ns[:2] + sub[:3]
        return proj_name[:3]

    @classmethod
    def deduce_python_version(cls) -> str:
        return ".".join(map(str, sys.version_info[:2]))

    def get_current_version_and_lines(
        self, data: tp.Optional[_ProtoNamespace] = None
    ) -> tp.Tuple[str, tp.List[str]]:
        data = data or self.auxe.python_project
        dir_ = data.module_dir
        init = self.auxcon_file.parent / dir_ / "__init__.py"
        if not init.exists():
            logger.warning(
                "%s/%s does not exist, retuning 'n/a' as version!", dir_, init.name
            )
            return "n/a", []

        with init.open("r", encoding="utf8") as f:
            lines = f.readlines()
            for line in lines:
                if "__version__" in line:
                    version = line.strip().split('"', 2)[1]
                    break
            else:
                raise RuntimeError(f"version not found in {init}")

        return version, lines

    def get_release_notes(
        self, auxd: tp.Optional[_ProtoNamespace] = None
    ) -> tp.Dict[str, str]:
        data = auxd or self.auxd
        dir_ = data.project.source_dir
        notes = self.auxcon_file.parent / dir_ / "release-notes.txt"
        release_note: tp.Dict[str, str] = {}

        if not notes.exists():
            logger.warning(
                "%s/%s does not exist, retuning an empty dict!", dir_, notes.name
            )
            return {}
        with notes.open("r", encoding="utf8") as f:
            for line in f.readlines():
                version, note = line.strip().split(" ", 1)
                release_note[version] = note

        return release_note

    def create_source(self) -> None:
        source_dir = Path(self.auxd.project.source_dir)
        module_dir = source_dir / self.auxd.project.name.replace(".", "/")
        tests_dir = source_dir / "tests"

        for path in [source_dir, module_dir, tests_dir]:
            path.mkdir(parents=True, exist_ok=True)

        init_file = module_dir / "__init__.py"
        rn_file = source_dir / "release-notes.txt"
        if not init_file.exists():
            with init_file.open("w", encoding="utf-8") as f:
                f.write('__version__ = "0.0.0"\n')
        if not rn_file.exists():
            with rn_file.open("w", encoding="utf-8") as f:
                f.write("0.0.0 unreleased\n")



# Copyright (c) 2021-2023 Mario S. Könz; License: MIT
from pathlib import Path

from .._util import LazyVersionStr
from ._02_base import BaseComponent


class MonotonicVersionMixin(BaseComponent):
    def bake(self) -> None:
        super().bake()
        self._raise_if_iv_older_than_luv()

    @property
    def _luv_file(self) -> Path:
        res: Path = self.target / "auxilium" / "last-used-version.txt"
        return res

    def _raise_if_iv_older_than_luv(self) -> None:
        # iv :installed version
        # luv : last used version
        inst_vers = str(LazyVersionStr())
        lu_vers = self._get_luv()

        inst_prefix = "v"
        if inst_vers.startswith(inst_prefix):
            inst_vers = inst_vers[1:]
        else:
            inst_prefix = ""
        lu_prefix = "v"
        if lu_vers.startswith(lu_prefix):
            lu_vers = lu_vers[1:]
        else:
            lu_prefix = ""

        iv_split = list(map(int, inst_vers.split(".")))
        luv_split = list(map(int, lu_vers.split(".")))
        if iv_split < luv_split:
            raise RuntimeError(
                f"""
                you are trying to run aux bake with version adaux=={inst_prefix+inst_vers},
                but this project has been baked with a newer version {lu_prefix+lu_vers}.
                Consider upgrading:  pip install adaux --upgrade
                """
            )

    def _get_luv(self) -> str:
        if self._luv_file.exists():
            with self._luv_file.open("r") as f:
                return f.readline().strip()
        inst_vers = str(LazyVersionStr())
        return inst_vers

    def writeout(self) -> None:
        super().writeout()
        self._luv_file.parent.mkdir(parents=True, exist_ok=True)
        with self._luv_file.open("w") as f:
            f.write(f"{LazyVersionStr()}\n")



# Copyright (c) 2021-2023 Mario S. Könz; License: MIT
import contextlib
import logging
import os
import shlex
import shutil
import subprocess
import typing as tp
from pathlib import Path

from .._logging import logger
from .._proto_namespace import _ProtoNamespace
from .._util import subprocess_run
from ._00_extra_level import ExtraLevel
from ._03_meta import MetaMixin

__all__ = ["MultiManageMixin"]


class MultiManageMixin(MetaMixin):
    @classmethod
    def __keys(cls) -> tp.Tuple[str, ...]:
        return ("projects", "tasks")

    def demodata(self) -> None:
        super().demodata()
        self.auxcon.multimanage = _ProtoNamespace(
            projects=["./admem"], tasks=_ProtoNamespace(ll=["ls -la"])
        )

    def formatted(self) -> None:
        super().formatted()
        self._copy_keys_over(self.__keys(), "multimanage")

        self._to_list("multimanage", "projects")
        self._to_proto_ns("multimanage", "tasks")

    def defaulted(self) -> None:
        super().defaulted()
        self.auxd.setdefault("multimanage", _ProtoNamespace())
        self.auxd.multimanage.setdefault("projects", [])
        self.auxd.multimanage.setdefault("tasks", _ProtoNamespace())

    def enriched(self) -> None:
        super().enriched()
        data = self.auxe.multimanage

        for project in data.projects:
            if not Path(project).exists():
                raise RuntimeError(f"project {project} does not exist")

        for _, task in data.tasks.items():
            assert isinstance(task, list)

    def mm_update(self, all_vip_branches: bool = False, prune: bool = False) -> None:
        def run(auxcon: _ProtoNamespace) -> None:
            branches = list(auxcon.gitlab["vip_branches"].keys())
            if not all_vip_branches:
                branches = branches[:1]
            prune_flag = ""
            if prune:
                prune_flag = "-p"
            self.subprocess_run_shlex(f"git fetch --all {prune_flag}".strip())
            # reversed, as we want to end up on default
            for branch in reversed(branches):
                self.subprocess_run_shlex(f"git checkout {branch}")
                self.subprocess_run_shlex("git pull")
            if prune:
                lbranches = self.subprocess_run_shlex_out(
                    "git branch --format='%(refname:short)'"
                ).split("\n")
                rbranches = self.subprocess_run_shlex_out(
                    "git branch -r --format='%(refname:short)'"
                ).split("\n")
                for lbra in lbranches:
                    if not any(lbra in rbra for rbra in rbranches):
                        ans = self._prompt(  # type: ignore
                            f"delete local branch '{lbra}' (y/n)", fg="red"
                        )
                        if ans not in "yY":
                            continue
                        self.subprocess_run_shlex(f"git branch -D {lbra}")

        summary = ["adaux mm update"]
        if all_vip_branches:
            summary.append("-a")
        if prune:
            summary.append("-p")

        self._run_with_report(run, " ".join(summary))

    def _run_with_report(
        self, run: tp.Callable[[_ProtoNamespace], None], summary_desc: str = ""
    ) -> None:
        report = []
        for project in self.auxe.multimanage.projects:
            auxcon = self.get_subproject(project, level=ExtraLevel.RAW)
            with self.project_header(auxcon), self.preserve_cwd(Path(project)):
                try:
                    run(auxcon)
                    report.append([auxcon.project.name, "pass"])
                except RuntimeError as err:
                    self._print(err.args[0], fg="red")
                    report.append([auxcon.project.name, "fail"])
                    continue

        with self.project_header(
            _ProtoNamespace(project=_ProtoNamespace(name=summary_desc))
        ):
            for name, state in report:
                col = ["red", "green"][state == "pass"]
                self._print(f"[{state}] {name}", fg=col)

    def get_subproject(
        self, project: str, level: ExtraLevel = ExtraLevel.ENRICHED
    ) -> _ProtoNamespace:
        with self.preserve_cwd():
            # pylint: disable=too-many-function-args,unexpected-keyword-arg
            x = self.__class__(Path(project), silent=True)  # type: ignore
            x.load_auxcon()
            with x.extra(level=level) as aux:  # type: ignore
                return aux  # type: ignore

    def mm_adaux(self, cmd_str: str) -> None:
        verbose = "-n"
        if logger.getEffectiveLevel() == logging.DEBUG:
            verbose = " -nv"
        self.mm_run(f"adaux {verbose} {cmd_str}")

    def mm_run(self, cmd_str: str) -> None:
        def run(auxcon: _ProtoNamespace) -> None:  # pylint: disable=unused-argument
            cmd_seq = self.auxe.multimanage.tasks.get(cmd_str, [cmd_str])
            for x in cmd_seq:
                self.subprocess_run_shlex(x, capture_output=False)

        self._run_with_report(run, f"{cmd_str}")

    @contextlib.contextmanager
    def project_header(self, auxcon: _ProtoNamespace) -> tp.Iterator[None]:
        columns, _ = shutil.get_terminal_size((80, 20))
        res = auxcon.project.name
        pad = columns - len(res) - 1
        self._print(res, fg="yellow", nl=False, bold=True)
        self._print(" " + pad * ">", fg="white")
        yield
        # self._print(pad * "<" + " " + res, fg="yellow")

    @contextlib.contextmanager
    def preserve_cwd(self, new_path: tp.Optional[Path] = None) -> tp.Iterator[None]:
        original_cwd = Path.cwd()
        if new_path:
            logger.debug("cwd changed to %s", new_path)
            os.chdir(new_path)
        try:
            yield
        finally:
            logger.debug("cwd changed to %s", original_cwd)
            os.chdir(original_cwd)

    def subprocess_run_shlex(
        self, cmd_str: str, *args: tp.Any, show: bool = True, **kwgs: tp.Any
    ) -> "subprocess.CompletedProcess[bytes]":
        kwgs.setdefault("check", True)
        kwgs.setdefault("capture_output", True)
        if show:
            columns, _ = shutil.get_terminal_size((80, 20))
            msg = f"> {cmd_str}"
            padding = columns - len(msg)
            self._print(msg, fg="blue", nl=not kwgs["capture_output"])
        cmd = shlex.split(cmd_str)
        try:
            res = subprocess_run(cmd, *args, **kwgs)
            if show and kwgs["capture_output"]:
                self._print((padding - 4) * " " + "[ok]", fg="green")
            return res
        except subprocess.CalledProcessError as err:
            if show and kwgs["capture_output"]:
                self._print((padding - 6) * " " + "[fail]", fg="red")
            raise RuntimeError(f"error when running {cmd}, see error above") from err

    def subprocess_run_shlex_out(self, *args: tp.Any, **kwgs: tp.Any) -> str:
        kwgs.setdefault("show", False)
        kwgs.setdefault("capture_output", True)
        if not kwgs["capture_output"]:
            raise RuntimeError("cannot use output_as and capture_output=False")

        res = self.subprocess_run_shlex(*args, **kwgs)

        return res.stdout.decode().strip()



# Copyright (c) 2021-2023 Mario S. Könz; License: MIT
import typing as tp

from .._logging import logger
from .._proto_namespace import _ProtoNamespace
from .._util import LazyVersionStr
from ._00_extra_level import ExtraLevel
from ._04_monotonic_version import MonotonicVersionMixin


class MigrateMixin(MonotonicVersionMixin):
    def __init__(self, *args: tp.Any, **kwgs: tp.Any) -> None:
        super().__init__(*args, **kwgs)
        self._migration_points = [
            "2.8.4",
            "3.0.0",
        ]

    @classmethod
    def __keys(cls) -> tp.Tuple[str, ...]:
        return ("version",)

    def formatted(self) -> None:
        super().formatted()
        self._copy_keys_over(self.__keys(), "meta", "migrate")

    def defaulted(self) -> None:
        super().defaulted()
        self.auxd.meta.setdefault("migrate", _ProtoNamespace())
        self.auxd.meta.migrate.setdefault("version", str(LazyVersionStr()))

    def migrate(self, target_version: tp.Optional[str] = None) -> None:
        self.auxcon.setdefault("meta", _ProtoNamespace())
        self.auxcon.meta.setdefault("migrate", _ProtoNamespace())
        from_version = self.auxcon.meta.migrate.get(
            "version", self._migration_points[0]
        )
        target_version = target_version or self._migration_points[-1]
        mig_path = []
        for to_version in self._migration_points:
            if from_version < to_version <= target_version:
                mig_path.append(to_version)
        if mig_path:
            logger.info("migration path is ->%s", "->".join(mig_path))
        else:
            logger.info("no migration necessary")

        for to_version in mig_path:
            self.step_migrate(to_version)

        self.auxcon.meta.migrate.version = to_version

        with self.extra(ExtraLevel.FORMATTED) as aux:  # type: ignore
            self.save_auxcon(aux)



# Copyright (c) 2021-2023 Mario S. Könz; License: MIT
import typing as tp

from .._proto_namespace import _ProtoNamespace
from ._06_dependency import DependencyMixin


class PipMixin(DependencyMixin):
    @classmethod
    def __keys(cls) -> tp.Tuple[str, ...]:
        return ("extra_index_url", "branch_match")

    def formatted(self) -> None:
        super().formatted()
        self._copy_keys_over(self.__keys(), "pip")
        self._to_proto_ns("pip", iter_mapping=True)

    def defaulted(self) -> None:
        super().defaulted()
        self.auxd.setdefault("pip", _ProtoNamespace())
        for key in self.__keys():
            self.auxd.pip.setdefault(key, _ProtoNamespace())

    def demodata(self) -> None:
        super().demodata()
        self.auxcon.setdefault("pip", _ProtoNamespace())
        data = self.auxcon.pip
        data.extra_index_url = _ProtoNamespace(
            demo="https://gitlab-ci-token:$CI_JOB_TOKEN@gitlab.x.y/api/v4/projects/118/packages/pypi/simple"
        )
        data.branch_match = _ProtoNamespace(
            demo=_ProtoNamespace(
                url="https://gitlab-ci-token:$CI_JOB_TOKEN@gitlab.x.y/user/demo.git",
                fallback="develop",
            )
        )

    def enriched(self) -> None:
        super().enriched()
        wo_creds = {}
        creds = {}

        data = self.auxe.pip
        for key, url in data.extra_index_url.items():
            if "@" in url:
                pre, url = url.split("@", 1)
                proto, cred = pre.split("//", 1)
                url = f"{proto}//{url}"
                token, var_cred = cred.split(":", 1)
                assert "$" == var_cred[0]
                creds[key] = var_cred[1:]
                assert "$" not in token

            wo_creds[key] = url

        data.extra_index_url_wo_creds = wo_creds
        data.creds = creds
        name = self.auxe.project.name
        version = self.auxe.python_project.version
        data.metadata_filecontent = (
            f"Metadata-Version: 2.1\\nName: {name}\\nVersion: {version}\\n"
        )

    def bake(self) -> None:
        super().bake()

        # we rely on .netrc for the credentials
        name = "pip/pip.conf"
        self.bake_file(name)

    def branch_match_and_cred_passing(
        self, opts: _ProtoNamespace, auxe: _ProtoNamespace
    ) -> None:
        opts.pip_extra_url = self._collect_extra_url(opts.pip_req, auxe)
        opts.pip_cred_vars = self._collect_cred_vars(opts.pip_req, auxe)
        for i, dep in enumerate(opts.pip_req):
            pkg, _ = self.parse_dep(dep)
            match = auxe.pip.branch_match.get(pkg)
            if match is None:
                continue
            # skip matching these branches
            skip = [auxe.gitlab.release_branch]
            var = f'{pkg.upper().replace(".", "_").replace("-", "_")}_MATCHING_REPO'
            opts.pip_req[i] = f"${{{var}:-{dep}}}"
            opts.pip_cred_vars.append(var)
            opts.setdefault("branch_match", [])
            opts.branch_match.append(
                (pkg, var, match["url"], skip, match.get("fallback"))
            )  # variable used by CI

    def _collect_extra_url(
        self, deps: tp.List[str], auxe: _ProtoNamespace
    ) -> tp.List[str]:
        extra_url = set()

        for dep in deps:
            pkg, _ = self.parse_dep(dep)
            url = auxe.pip.extra_index_url.get(pkg)
            if url is not None:
                extra_url.add(url)

        return list(sorted(extra_url))

    def _collect_cred_vars(
        self, deps: tp.List[str], auxe: _ProtoNamespace
    ) -> tp.List[str]:
        cred_vars = set()

        for dep in deps:
            pkg, _ = self.parse_dep(dep)
            cred = auxe.pip.creds.get(pkg)
            if cred is not None:
                cred_vars.add(cred)

        return list(sorted(cred_vars))



# Copyright (c) 2021-2023 Mario S. Könz; License: MIT
import collections
import typing as tp

from .._proto_namespace import _ProtoNamespace
from ._02_base import BaseComponent


class MetaMixin(BaseComponent):
    def formatted(self) -> None:
        super().formatted()
        keys = ["disable", "enable", "migrate"]
        self._copy_keys_over(keys, "meta")
        self._to_list("meta", "disable")
        self._to_list("meta", "enable")
        # migrate is handeled by MigrateMixin

    def defaulted(self) -> None:
        super().defaulted()
        self.auxd.setdefault("meta", _ProtoNamespace())
        self.auxd.meta.setdefault("disable", [])
        self.auxd.meta.setdefault("enable", [])

        if self.auxd.meta.disable and self.auxd.meta.enable:
            self._error_both()

    def _error_both(self) -> None:
        raise RuntimeError("Cannot set enable and disable at the same time.")

    def demodata(self) -> None:
        super().demodata()
        self.auxcon.meta = _ProtoNamespace(disable=["docs"])

    def is_enabled(self, component_name: str) -> bool:
        for x in self.__class__.__mro__:
            if self._comp_name_type_match(component_name, x):
                return self._component_enabled(component_name)
        return False

    def type_wo_disabled(
        self,
        *,
        discard_before: str = "",
        check_absence: bool = True,
    ) -> "tp.Type[MetaMixin]":
        res: tp.List[type] = []
        keep_after = "MetaMixin"
        for part in self.__class__.__mro__:
            if part.__name__ == keep_after:
                res.append(part)
                continue
            if part.__name__ == discard_before:
                res.clear()
            if part.__name__ in ["AllComponents", "DynComponent"]:
                continue
            if self._part_enabled(part):
                res.append(part)

        res_type = BaseComponent.compose(*reversed(res))

        # check if disabled did not get added by enabled
        if self._uses_disable() and check_absence:
            compare_to = self.type_wo_disabled(
                discard_before="SentinelMixin", check_absence=False
            )
            parents: tp.MutableMapping[
                type, tp.Sequence[type]
            ] = collections.defaultdict(list)
            for part in compare_to.__mro__[1:]:  # remove bottom dyn
                parents[part] = part.__mro__[1:]  # remove self == part
                if not self._part_enabled(part):
                    used_by = [
                        key.__name__ for key, val in parents.items() if part in val
                    ]
                    used_by2 = [val for key, val in parents.items() if part in val]
                    print(used_by2)
                    raise RuntimeError(
                        f"{part.__name__} cannot be disabled, as it is used by '{', '.join(used_by)}'"
                    )

        return res_type  # type: ignore

    def _en_disabled_list(self) -> tp.Tuple[tp.List[str], tp.List[str]]:
        disable = self.auxcon.get("meta", _ProtoNamespace()).get("disable", [])
        enable = self.auxcon.get("meta", _ProtoNamespace()).get("enable", [])
        return enable, disable

    def _component_enabled(self, component_name: str) -> bool:
        enable, disable = self._en_disabled_list()
        if enable:
            if disable:
                self._error_both()
            return component_name in enable
        if disable:
            return component_name not in disable
        return True

    def _uses_disable(self) -> bool:
        enable, disable = self._en_disabled_list()
        if enable:
            if disable:
                self._error_both()
            return False
        return True

    def _part_enabled(self, part: tp.Type[tp.Any]) -> bool:
        enable, disable = self._en_disabled_list()
        if enable:
            if disable:
                self._error_both()
            if part in MetaMixin.__mro__:  # keep before (and)
                return True
            # see SentinelMixin for other end
            return any(self._comp_name_type_match(x, part) for x in enable)
        if disable:
            return not any(self._comp_name_type_match(x, part) for x in disable)
        return True

    @classmethod
    def _comp_name_type_match(
        cls, component_name: str, component_type: tp.Type[tp.Any]
    ) -> bool:
        return cls._canon_comp_name(component_name) == cls._canon_type_name(
            component_type
        )

    @classmethod
    def _canon_type_name(cls, component_type: tp.Type[tp.Any]) -> str:
        return component_type.__name__.lower().replace("mixin", "")

    @classmethod
    def _canon_comp_name(cls, component_name: str) -> str:
        return component_name.replace("-", "")

    def step_migrate(self, to_version: str) -> None:
        super().step_migrate(to_version)

        if to_version == "3.0.0":
            if "meta" in self.auxcon and "disabled" in self.auxcon.meta:
                self.auxcon.meta.disable = self.auxcon.meta.pop("disabled")



# Copyright (c) 2021-2023 Mario S. Könz; License: MIT
import logging

import colorlog

logger = logging.getLogger("adaux")

handler = logging.StreamHandler()
handler.setLevel(logging.DEBUG)
formatter = colorlog.ColoredFormatter(
    "%(log_color)s%(levelname)-8s%(reset)s %(blue)s%(message)s",
    datefmt=None,
    reset=True,
    log_colors={
        "DEBUG": "cyan",
        "INFO": "green",
        "WARNING": "yellow",
        "ERROR": "red",
        "CRITICAL": "red,bg_white",
    },
    secondary_log_colors={},
    style="%",
)

handler.setFormatter(formatter)
logger.addHandler(handler)



# Copyright (c) 2021-2023 Mario S. Könz; License: MIT
from . import _cli
from . import _cli_mixin
from . import _components
from . import _parser
from ._cli import *
from ._cli_gitlab import *
from ._cli_multimanage import *
from ._components import _00_extra_level
from ._components._00_extra_level import *
from ._parser import *

__all__ = _cli.__all__ + _00_extra_level.__all__ + _parser.__all__

__version__ = "3.2.0"



# pylint: disable=relative-beyond-top-level
import os
import typing as tp

from ...._components._payload._docker_executors import subprocess_run
from ...._components._payload._docker_executors import tag_image
from ...._components._payload._docker_executors import upload_to_remote
from ...._proto_namespace import _ProtoNamespace


def check_release_notes(aux: _ProtoNamespace) -> None:
    release_notes = aux.python_project.release_notes
    version = aux.python_project.version
    # check that version has a note
    if version not in release_notes:
        raise RuntimeError(f"version {version} is not in release notes, please add!")
    # check that version was not already released
    if version in aux.env_var._remote_version_tags:  # pylint: disable=protected-access
        raise RuntimeError(
            f"version {version} was already released and cannot be released again!"
        )

    print(f"version {version} has description '{release_notes[version]}'")


def env_variables(auxh: tp.Any, **kwgs: str) -> None:
    for key, val in kwgs.items():
        val = auxh.env_var.apply_format(val)
        # this is acutally persistent, as we run in the
        # same python process as the rest of adaux.
        # (would not be possible with subshells)
        os.environ[key] = val
        print(f"{key}={val}")


def tag(
    auxh: _ProtoNamespace,
    deps: tp.Any,
    tags: tp.Union[str, tp.Sequence[str]],  # pylint: disable=redefined-outer-name
) -> None:
    if isinstance(tags, str):
        tags = [tags]

    last_local_tag = ""
    for dep in deps:
        for tag_ in tags:
            tag_ = auxh.env_var.apply_format(tag_)
            local_tag, release_tag = dep.executor.tag_and_upload(tag_)
            if last_local_tag != local_tag:
                last_local_tag = local_tag
                msg = "uploaded" if dep.executor.remote_exists() else "  tagged"
                print(msg, local_tag)
            print("   -> to", release_tag)


def gittag(
    auxh: _ProtoNamespace,
    deps: tp.Any,
    tags: tp.Union[str, tp.Sequence[str], None] = None,
) -> None:
    if tags is None:
        # get it from the deps
        assert len(deps) == 1
        dep = deps[0]
        tags = dep.param["tags"]

    if isinstance(tags, str):
        tags = [tags]

    for tag_ in tags:
        tag_ = auxh.env_var.apply_format(tag_)
        subprocess_run(["git", "tag", "-f", tag_])
        url = os.environ["CI_REPOSITORY_URL"]
        protocol, rest = url.split("//", 1)
        _, rest = rest.split("@", 1)
        url = f"{protocol}//__token__:{os.environ['GITLAB_WRITE_REPOSITORY']}@{rest}"
        subprocess_run(["git", "push", url, tag_], check=True)


def img_dockerhub(
    auxh: _ProtoNamespace,
    deps: tp.Any,
    release_tag: str,
) -> None:
    if len(deps) != 1:
        raise RuntimeError(
            f"img-dockerhub job for {release_tag} should have exactly 1 dependency, not {len(deps)}!"
        )
    release_tag = auxh.env_var.apply_format(release_tag)
    local_tag = deps[0].executor.pull_if_not_existent()
    tag_image(local_tag, release_tag)
    subprocess_run(
        [
            "docker",
            "login",
            "-u",
            os.environ["DOCKERHUB_USERNAME"],
            "-p",
            os.environ["DOCKERHUB_PASSWORD"],
            "docker.io",
        ]
    )
    upload_to_remote(local_tag, release_tag)
    subprocess_run(["docker", "logout", "docker.io"])
    print("uploaded", local_tag)
    print("   -> to", release_tag)
